import bpy
import os
import sys
from pathlib import Path
from math import degrees
import bmesh


# print('starteddddddddddddddddddd')
# for i, arg in enumerate(sys.argv):
#     print('aaaaaaaaaaarg:', arg)

blender_executable, blend_file = sys.argv[:2]
script_file = sys.argv[4]
out_file_path = sys.argv[5]

scene_name = Path(bpy.data.filepath).stem
print('file_name:', scene_name)

code = ''

if not '--models_only' in sys.argv:
    code += '''from ursina import *
from time import perf_counter

scene_parent = Entity()

if __name__ == '__main__':
    app = Ursina()

t = perf_counter()
'''
code += '''
# unique meshes
meshes = {
'''

dg = bpy.context.evaluated_depsgraph_get() #getting the dependency graph
bpy.ops.object.mode_set(mode='OBJECT', toggle=False)
objects = [ob for ob in bpy.data.objects if ob.type == 'MESH']
unique_objects = {}
# print('meshes:', meshes)
for ob in objects:
    unique_objects[ob.data.name] = ob


for key, ob in unique_objects.items():
    mesh = ob.evaluated_get(dg).data

    # triangulate mesh
    bm = bmesh.new()
    bm.from_mesh(mesh)
    bmesh.ops.triangulate(bm, faces=bm.faces[:], quad_method='BEAUTY', ngon_method='BEAUTY')
    bm.normal_update()
    bm.to_mesh(mesh)
    bm.free()

    verts = []
    normals = []
    vertex_colors = []
    uvs = []

    for poly in mesh.polygons:
        face_normal_x, face_normal_y, face_normal_z = 0, 0, 0

        for idx in poly.vertices:
            v = mesh.vertices[idx]
            verts.append((round(v.co[0],4),round(v.co[2],4),round(v.co[1],4)))
            normals.append((v.normal[0], v.normal[2], v.normal[1]))

    # average the vertex normals ot get face normals
    sharp_normals = []
    for i in range(0, len(normals), 3):
        averaged_normal = (
            round(sum((normals[i][0], normals[i+1][0], normals[i+2][0])) / 3, 5),
            round(sum((normals[i][1], normals[i+1][1], normals[i+2][1])) / 3, 5),
            round(sum((normals[i][2], normals[i+1][2], normals[i+2][2])) / 3, 5),
            )
        for j in range(3):
            sharp_normals.append(averaged_normal)

    normals = sharp_normals

    color_layer = mesh.vertex_colors.active
    if color_layer and len(color_layer.data) > 0:
        vertex_colors = [tuple(round(e, 4) for e in data.color) for data in color_layer.data]


    uv_layer = mesh.uv_layers.active
    if uv_layer and len(uv_layer.data) > 0:
        uvs = [tuple(round(e, 4) for e in data.uv) for data in uv_layer.data]



    code += f'''
'{mesh.name.replace('.', '_')}' : Mesh(
    vertices={verts},
    normals={normals},
    colors={vertex_colors},
    uvs={uvs},
),
'''

code += '}\n'

if '--models_only' in sys.argv:
    print('sucessfully exported blender models')
    with open(out_file_path, 'w') as f:
        f.write(code)

else:
    # write entities
    code += '''print('loaded models:', perf_counter() - t)\n'''
    code += '''t = perf_counter()\n'''


    for ob in objects:
        if '--skip-hidden' in sys.argv and ob.hide_get() == True:
            continue

        #deselect all but just one object and make it active
        bpy.ops.object.select_all(action='DESELECT')
        rotation = [degrees(e) for e in ob.matrix_world.to_euler()]

        code += f'''
scene_parent.{ob.name.replace('.', '_')} = Entity(
    name=\'{ob.name}\',
    parent=scene_parent,
    position=Vec3({round(ob.location.x,5)}, {round(ob.location.z,5)}, {round(ob.location.y,5)}),
    rotation=({round(-rotation[0],5)}, {round(-rotation[2],5)}, {round(rotation[1],5)}),
    scale=Vec3({round(ob.scale.x,5)}, {round(ob.scale.z,5)}, {round(ob.scale.y,5)}),
    '''

        code += f'''model=deepcopy(meshes['{ob.data.name.replace('.', '_')}']),'''
        # code += f'''model='cube','''

        if ob.active_material:
            color = ob.active_material.diffuse_color
            code += f'''
    color=({round(color[0],5)}, {round(color[1],5)}, {round(color[2],5)}, {round(color[3],5)}),'''

        code += f'''
    ignore=True,
    )'''

    code += '''\nprint('created entities:', perf_counter() - t)'''
    #        if ob.parent:
    #            code += f'''parent=self.{ob.name.replace('.', '_')}'''



        #This has to be done every time the object updates:
        # ob = ob.evaluated_get(dg) #this gives us the evaluated version of the object. Aka with all modifiers and deformations applied.
        # mesh = ob.to_mesh() #turn it into the mesh data block we want.
        # verts = []
        # tris = []
        # uvs = []
        # normals = []
        # vertex_colors = []

        # for poly in mesh.polygons:
        #     tris.append(tuple(e for e in poly.vertices))
        # # for poly in mesh.polygons:
        # #     for idx in poly.vertices:
        # #         verts.append(mesh.vertices[idx])
        # #         normal = mesh.vertices[idx].normal
        # #         n = (normal[0], normal[1], normal[2])
        # #         normals.append(n)
        #
        #
        # color_layer = mesh.vertex_colors.active
        # if color_layer and len(color_layer.data) > 0:
        #     if not vertex_colors:
        #         vertex_colors = [(1,1,1,1) for e in mesh.vertices]
        #
        #     i = 0
        #     for poly in mesh.polygons:
        #         for idx in poly.vertices:
        #             vertex_colors[idx] = tuple(round(e, 4) for e in color_layer.data[i].color)
        #             i += 1
        #
        # uv_layer = mesh.uv_layers.active
        # if uv_layer and len(uv_layer.data) > 0:
        #     if not uvs:
        #         uvs = [(0,0) for e in mesh.vertices]
        #
        #     i = 0
        #     for poly in mesh.polygons:
        #         for idx in poly.vertices:
        #             uvs[idx] = tuple(round(e, 4) for e in uv_layer.data[i].uv)
        #             i += 1


    #     code += f'''
    # model=Mesh('''
    #     # code += f'''
    #     # vertices=[{', '.join([f'Vec3({v.co[0]}, {v.co[2]}, {v.co[1]})' for v in verts])}],'''
    #
    #     code += f'''
    #     vertices=[{', '.join([f'Vec3({round(v.co[0], 4)}, {round(v.co[2], 4)}, {round(v.co[1], 4)})' for v in mesh.vertices])}],
    #     triangles={tris},'''
    #
    #     code += f'''
    #     normals=({' ,'.join([f'Vec3({round(v.normal[0], 4)}, {round(v.normal[2], 4)}, {round(v.normal[1], 4)})' for v in mesh.vertices])}),'''
    #     # code += f'''
    #     # normals=({' ,'.join([f'Vec3({str(n[0])[5:]}, {n[2]}, {n[1]})' for v in normals])}),'''
    #
    #     if uvs:
    #         code += f'''
    #     uvs={uvs},'''
    #
    #     if vertex_colors:
    #         code += f'''
    #     colors={vertex_colors},'''
    #
    #     code += '\n    )'
    #     code += '\n)'



    code += '''
if __name__ == '__main__':
    EditorCamera()
    app.run()
'''

    # return code
    print('sucessfully exported blender scene')
    with open(out_file_path, 'w') as f:
        f.write(code)



import bpy
import os
import sys


blend_file_path = bpy.data.filepath
filepath = sys.argv[-1]


bpy.ops.export_scene.obj(
    filepath=filepath,
    check_existing=True,
    axis_forward='Z',
    axis_up='Y',
    filter_glob="*.obj;*.mtl",
    use_selection=False,
    use_animation=False,
    use_mesh_modifiers=True,
    use_edges=False,
    use_smooth_groups=False,
    use_smooth_groups_bitflags=False,
    use_normals=True,
    use_uvs=True,
    use_materials=False,
    use_triangles=True,
    use_nurbs=False,
    use_vertex_groups=False,
    use_blen_objects=True,
    group_by_object=False,
    group_by_material=False,
    keep_vertex_order=False,
    global_scale=1,
    path_mode='AUTO'
)



import sys
from ursina.entity import Entity
from panda3d.core import PerspectiveLens, OrthographicLens, LensNode, NodePath
from panda3d.core import Camera as PandaCamera
from panda3d.core import Texture as PandaTexture
from direct.filter.FilterManager import FilterManager
from ursina import application
from ursina.scene import instance as scene
from ursina.window import instance as window
from ursina import color
from ursina.texture import Texture
from ursina.shader import Shader



class Camera(Entity):

    def __init__(self):
        super().__init__()
        self.parent = scene
        self.name = 'camera'
        self.eternal = True

        self._cam = None
        self._render = None
        self.ui_size = 40
        self._ui_lens_node = None
        self.ui = None
        self.fov = 40
        self.orthographic = False


    def set_up(self):
        self.display_region = base.camNode.get_display_region(0)
        win = self.display_region.get_window()

        self.perspective_lens = PerspectiveLens()
        self.perspective_lens = base.camLens # use panda3d's default for automatic aspect ratio on window resize
        self.lens = self.perspective_lens
        self.perspective_lens.set_aspect_ratio(window.aspect_ratio)
        self.perspective_lens_node = LensNode('perspective_lens_node', self.perspective_lens)
        self.lens_node = self.perspective_lens_node

        self.orthographic_lens = OrthographicLens()
        self.orthographic_lens.set_film_size(self.fov * window.aspect_ratio, self.fov)
        self.orthographic_lens_node = LensNode('orthographic_lens_node', self.orthographic_lens)

        application.base.cam.node().set_lens(self.lens)

        self.orthographic = False
        self.fov = 40
        self.clip_plane_near = 0.1
        self.clip_plane_far = 10000

        self.ui_display_region = win.make_display_region()
        self.ui_display_region.set_sort(20)

        self.ui_camera = NodePath(PandaCamera('ui_camera'))
        self.ui_lens = OrthographicLens()
        # moved set_film_size() to window module for correct aspect ratio after setting window size
        self.ui_lens.set_near_far(-1000, 1000)
        self.ui_camera.node().set_lens(self.ui_lens)
        self._ui_lens_node = LensNode('_ui_lens_node', self.ui_lens)

        self.ui_render = NodePath('ui_render')
        self.ui_render.set_depth_test(0)
        self.ui_render.set_depth_write(0)
        self.ui_camera.reparent_to(self.ui_render)
        self.ui_display_region.set_camera(self.ui_camera)
        scene.ui_camera = self.ui_camera

        self.ui = Entity(eternal=True, name='ui', parent=self.ui_camera, scale=(self.ui_size*.5, self.ui_size*.5))
        self.overlay = Entity(parent=self.ui, model='quad', scale_x=self.aspect_ratio, color=color.clear, eternal=True, z=-99)

        # these get created when setting a shader
        self.filter_manager = None
        self.filter_quad = None
        self.render_texture = None
        self.filter_quad = None
        self.depth_texture = None
        # self.normals_texture = None


    @property
    def orthographic(self):
        return self._orthographic

    @orthographic.setter
    def orthographic(self, value):
        self._orthographic = value
        self.lens_node = (self.perspective_lens_node, self.orthographic_lens_node)[value] # this need to be set for the mouse raycasting
        application.base.cam.node().set_lens((self.perspective_lens, self.orthographic_lens)[value])
        self.fov = self.fov

    @property
    def fov(self):
        return self._fov

    @fov.setter
    def fov(self, value):
        value = max(1, value)
        self._fov = value
        if not self.orthographic and hasattr(self, 'perspective_lens'):
            self.perspective_lens.set_fov(value)

        elif self.orthographic and hasattr(self, 'orthographic_lens'):
            self.orthographic_lens.set_film_size(value * self.aspect_ratio, value)

        application.base.cam.node().set_lens((self.perspective_lens, self.orthographic_lens)[value])

    @property
    def clip_plane_near(self):
        return self.lens.getNear()

    @clip_plane_near.setter
    def clip_plane_near(self, value):
        self.lens.set_near(value)

    @property
    def clip_plane_far(self):
        return self.lens.getFar()

    @clip_plane_far.setter
    def clip_plane_far(self, value):
        self.lens.set_far(value)

    @property
    def aspect_ratio(self):
        return self.perspective_lens.get_aspect_ratio()

    @property
    def shader(self):
        return self._shader

    @shader.setter
    def shader(self, value):
        self._shader = value
        if value is None:
            self.filter_manager.cleanup()
            self.filter_manager = None
            if self.filter_quad:
                self.filter_quad.removeNode()
                # print('removed shader')
            return None

        shader = value

        if isinstance(value, Shader) and not shader.compiled:
            shader.compile()

        if isinstance(value, Shader):
            shader = shader._shader

        if not self.filter_manager:
            self.filter_manager = FilterManager(base.win, base.cam)
            self.render_texture = PandaTexture()
            self.depth_texture = PandaTexture()
            # self.normals_texture = PandaTexture()
            # from panda3d.core import AuxBitplaneAttrib
            self.filter_quad = self.filter_manager.renderSceneInto(
                colortex=self.render_texture,
                depthtex=self.depth_texture,
                # auxtex=self.normals_texture,
                # auxbits=AuxBitplaneAttrib.ABOAuxNormal
                )
            self.filter_quad.set_shader_input("tex", self.render_texture)
            self.filter_quad.set_shader_input("dtex", self.depth_texture)
            # self.filter_quad.set_shader_input("ntex", self.normals_texture)

        self.filter_quad.setShader(shader)

        if hasattr(value, 'default_input'):
            for key, value in value.default_input.items():
                if callable(value):
                    value = value()

                self.set_shader_input(key, value)


        print('set camera shader to:', shader)


    def set_shader_input(self, name, value):
        if self.filter_quad:

            if isinstance(value, Texture):
                value = value._texture    # make sure to send the panda3d texture to the shader

            self.filter_quad.setShaderInput(name, value)
        else:
            print('no filter quad')


instance = Camera()


if __name__ == '__main__':
    from ursina import *
    window.borderless = False
    app = Ursina()

    camera.orthographic = True

    e = Entity()
    e.model = 'quad'
    e.color = color.random_color()
    e.position = (-2, 0, 10)

    e = Entity()
    e.model = 'quad'
    e.color = color.random_color()
    e.position = (2, 0, 10)

    e = Entity()
    e.model = 'quad'
    e.color = color.random_color()
    e.position = (0, 0, 40)

    EditorCamera()
    # from ursina import *
    # Button(text='test button')
    from ursina.shaders import camera_grayscale_shader
    camera.shader = camera_grayscale_shader

    # def update():
    #     t = Texture(camera.render_texture)
    #     print(t.pixels)
    app.run()



from ursina import *


class Animation(Sprite):
    def __init__(self, name, fps=12, loop=True, autoplay=True, frame_times=None, **kwargs):

        if isinstance(name, Path):
            texture = load_texture(name.name, name.parent)
        else:
            texture = load_texture(name)

        if texture and texture.path.suffix == '.gif':   # load gif
            import imageio
            from PIL import Image
            path = load_texture(name).path
            gif = imageio.get_reader(path)
            img = Image.open(path)
            img.seek(0)
            frame_times = []
            for i in range(len(gif)):
                img.seek(i)
                frame_times.append(img.info['duration'] / 1000)

            self.frames = [Texture(Image.fromarray(frame)) for frame in gif]

        else:   # load image sequence
            texture_folders = (application.compressed_textures_folder, application.asset_folder, application.internal_textures_folder)
            self.frames = [Texture(e) for e in find_sequence(name, ('png', 'jpg'), texture_folders)]


        if self.frames:
            super().__init__(texture=self.frames[0])

        self.sequence = Sequence(loop=loop, auto_destroy=False)

        self.frame_times = frame_times
        if not self.frame_times:
            self.frame_times = [1/fps for i in range(len(self.frames))]

        for i, frame in enumerate(self.frames):
            self.sequence.append(Func(setattr, self, 'texture', self.frames[i]))
            self.sequence.append(Wait(self.frame_times[i]))

        self.is_playing = False
        self.autoplay = autoplay


        for key, value in kwargs.items():
            setattr(self, key ,value)


        if self.autoplay:
            self.start()


    def start(self):
        if self.is_playing:
            self.finish()
        self.sequence.start()
        self.is_playing = True

    def pause(self):
        self.sequence.pause()

    def resume(self):
        self.sequence.resume()

    def finish(self):
        self.sequence.finish()
        self.is_playing = False


    @property
    def duration(self):     # get the duration of the animation. you can't set it. to do so, change the fps instead.
        return self.sequence.duration


    def __setattr__(self, name, value):
        if hasattr(self, 'frames') and name in ('color', 'origin'):
            for f in self.frames:
                setattr(f, name, value)

        if name == 'loop':
            self.sequence.loop = value

        try:
            super().__setattr__(name, value)
        except Exception as e:
            return e





if __name__ == '__main__':
    # application.asset_folder = application.asset_folder.parent.parent / 'samples'
    app = Ursina()

    '''
    Loads an image sequence as a frame animation.
    So if you have some frames named image_000.png, image_001.png, image_002.png and so on,
    you can load it like this: Animation('image')

    You can also load a .gif by including the file type: Animation('image.gif')
    '''

    Animation('ursina_wink')
    # Animation('city_in_desert_valley_water.gif')

    app.run()



from ursina import *
import pyperclip
# from tree_view import TreeView


class TextField(Entity):
    def __init__(self, **kwargs):
        super().__init__(
            parent=camera.ui,
            x=-.5,
            y=.4,
            ignore_paused=True,
            )


        self.font = 'VeraMono.ttf'
        self.line_height = 1
        # self.max_width = 80
        self.max_lines = 99999

        self.text_entity = Text(
            parent = self,
            start_tag='☾',
            end_tag = '☽',
            font = self.font,
            text='',
            line_height=self.line_height,
            origin=(-.5, .5),
            )
        self.line_numbers = Text(
            parent = self,
            font = self.font,
            text='0',
            origin=(.5,.5),
            x=-.04,
            color=color.gray,
            enabled = False,
            )
        self.character_width = Text.get_width('a', font=self.font)
        self.cursor_parent = Entity(parent=self, scale=(self.character_width, -1*Text.size))
        # self.max_line_indicatior = Entity(parent=self.cursor_parent, model='quad', origin=(-.5,.5), scale=(100,.05), rotation_x=180, color=color.red)
        # self.max_width_indicatior = Entity(
        #     parent=self.cursor_parent, model='quad', origin=(-.5,.5), scale=(100,.05), rotation_x=180, rotation_z=90, color=color.color(0,0,1,.05), x=80)
        self.cursor = Entity(parent=self.cursor_parent, model='cube', color=color.white33, origin=(-.5, -.5), scale=(.1, 1))
        self.bg = Entity(parent=self.cursor_parent, model='cube', color=color.dark_gray, origin=(-.5,-.5), z=1, scale=(120, 20), collider='box', visible=False)

        self.selection = None
        self.selection_parent = Entity(parent=self.cursor_parent)
        self.register_mouse_input = False

        def blink_cursor():
            if self.cursor.color == color.cyan:
                self.cursor.color = color.clear
            else:
                self.cursor.color = color.cyan
            invoke(blink_cursor, delay=.5)

        blink_cursor()
        self.text = ''

        self.replacements = dict()

        self.on_undo = list()
        self.on_redo = list()

        self.shifted_keys = {
            '-' : '_',
            '.' : ':',
            ',' : ';',
            '\'' : '*',
            '<' : '>',
            '+' : '?',
            '0' : '=',
            '1' : '!',
            '2' : '"',
            '3' : '#',
            # '4' : '¤',
            '5' : '%',
            '6' : '&',
            '7' : '/',
            '8' : '(',
            '9' : ')',
        }
        self.alted_keys = {
            '\'' : '´',
            '0' : '}',
            '2' : '@',
            '3' : '£',
            '4' : '¤',
            '5' : '€',
            '7' : '{',
            '8' : '[',
            '9' : ']',
        }
        self.shortcuts = {
            'newline':          ('enter', 'enter hold'),
            'erase':            ('backspace', 'backspace hold'),
            'erase_word':       ('ctrl+backspace', 'ctrl+backspace hold'),
            'delete_line':      ('ctrl+shift+k',),
            'undo':             ('ctrl+z', 'ctrl+z hold'),
            'redo':             ('ctrl+y', 'ctrl+y hold', 'ctrl+shift+z', 'ctrl+shift+z hold'),
            # 'save':             ('ctrl+s',),
            # 'save_as':          ('ctrl+shift+s',),
            'indent':           ('tab',),
            'dedent':           ('shift+tab',),
            'move_line_down':   ('ctrl+down arrow', 'ctrl+down arrow hold'),
            'move_line_up':     ('ctrl+up arrow', 'ctrl+up arrow hold'),
            # 'cut':              ('ctrl+x',),
            'copy':             ('ctrl+c',),
            'paste':            ('ctrl+v',),
            # 'select_all':       ('ctrl+a',),
            # 'toggle_comment':   ('ctrl+alt+c',),
            # 'find':             ('ctrl+f',),

            'move_left':        ('left arrow', 'left arrow hold'),
            'move_right':       ('right arrow', 'right arrow hold'),
            'move_up':          ('up arrow', 'up arrow hold'),
            'move_down':        ('down arrow', 'down arrow hold'),
            'move_to_end_of_word' : ('ctrl+right arrow', 'ctrl+right arrow hold'),
            'move_to_start_of_word' : ('ctrl+left arrow', 'ctrl+left arrow hold'),

            'select_word_left': ('ctrl+shift+left arrow', 'ctrl+shift+left arrow hold'),
            'select_word':      ('double click'),
        }

        # self.debug_cursor = Entity(parent=self.cursor_parent, model='cube', origin=(-.5,-.5), color=color.white33)


        for key, value in kwargs.items():
            setattr(self, key, value)

    def add_text(self, s, move_cursor=True):
        x, y = int(self.cursor.x), int(self.cursor.y)

        lines = self.text.split('\n')
        l = lines[y]
        lines[y] =  l[:x] + s + l[x:]
        self.on_undo.append((self.text, y, x))
        self.text = '\n'.join(lines)

        if move_cursor:
            self.cursor.x += len(s)


    def move_line(self, a, b):
        x, y = int(self.cursor.x), int(self.cursor.y)

        lines = self.text.split('\n')
        lines[a], lines[b] = lines[b], lines[a]
        self.on_undo.append((self.text, y, x))
        self.text = '\n'.join(lines)
        # print('moved line')


    def erase(self):
        # if not self.selection or self.selection[0] == self.selection[1]:
        x, y = int(self.cursor.x), int(self.cursor.y)
        if x+y == 0:
            return

        lines = self.text.split('\n')
        l = lines[y]
        # delete \n and go to line above
        if x == 0 and y > 0:
            new_x = len(lines[y-1])
            lines[y-1] += l
            lines.pop(y)
            self.cursor.x = new_x
            self.cursor.y -= 1
        # delete tab
        elif l[:x].lstrip() == '' and x >=4:
            l = l[4:]
            self.cursor.x -= 4
            lines[y] = l
        # normal erase
        else:
            removed = l[x-1]

            l = l[:x-1] + l[x:]
            self.cursor.x -= 1

            lines[y] = l

        self.on_undo.append((self.text, y, x))

        # else:   # delete selected
        #     self.delete_selected()
        #     lines = self.text.split('\n')
        #
        #     start_y = int(self.selection[0][1])
        #     end_y = int(self.selection[1][1])
        #
        #     if start_y == end_y:    # one line
        #         l = lines[start_y]
        #         start_x, end_x = int(self.selection[0][0]), int(self.selection[1][0])
        #         l = l[:start_x] + l[end_x:]
        #         lines[start_y] = l
        #         self.selection = (self.selection[0], self.selection[0])
        #         self.cursor.position = self.selection[0][0]
                # self.selection = None
                # self.selection[0][0],start_y),
                # ignore=True, scale_x=self.selection[1][0]-self.selection[0][0])
            # for y in range(int(self.selection[0][1]), int(self.selection[1][1])):
            #     lines.pop(y)




        self.text = '\n'.join(lines)

    def delete_selected(self):
        lines = self.text.split('\n')
        self.cursor.position = self.selection[1]
        if int(self.selection[1][1]) > int(self.selection[0][1]):
            for y in range(int(self.selection[0][1]), int(self.selection[1][1])):
                lines.pop(y)
                print('delete line:', y)


        for x in range(int(self.selection[1][0]) - int(self.selection[0][0])):
            self.erase()
        # l = lines[int(self.selection[1])]

        self.text = '\n'.join(lines)
        self.selection = None
        # self.on_undo.append((self.text, y, x))
        self.draw_selection()


    def input(self, key):
        text, cursor, on_undo, add_text, erase = self.text, self.cursor, self.on_undo, self.add_text, self.erase

        if key == 'space':
            key = ' '
        # key = f'{"ctrl" if held_keys["control"]} + {key}'
        ctrl, shift, alt = '', '', ''
        if held_keys['control'] and key != 'control':
            ctrl = 'ctrl+'
        if held_keys['shift'] and key != 'shift':
            shift = 'shift+'
        if held_keys['alt'] and key != 'alt':
            alt = 'alt+'

        key = ctrl+shift+alt+key
        # print(key)
        x, y = int(cursor.x), int(cursor.y)
        lines = text.split('\n')
        l = lines[y]

        if key in self.shortcuts['move_up']:
            if y > 0:
                cursor.y -= 1
                cursor.x = min(x, len(lines[y-1]))
        if key in self.shortcuts['move_down']:
            if y < len(lines)-1:
                cursor.y += 1
                cursor.x = min(x, len(lines[y+1]))
        if key in self.shortcuts['move_right']:
            if x == len(l) and y < len(lines)-1:        # end of line, move to beginning of next
                cursor.y += 1
                cursor.x = 0
            elif x < len(l):
                cursor.x += 1
        if key in self.shortcuts['move_left']:
            if x > 0:
                cursor.x -= 1
            elif y > 0:                                 # move to end of line above
                cursor.y -= 1
                cursor.x = len(lines[y-1])

        delimiters = (' ', '.')
        if key in self.shortcuts['move_to_end_of_word']:
            if x == len(l):
                if y < len(lines)-1:        # end of line, move to beginning of next
                    cursor.y += 1
                    cursor.x = 0
                return

            elif l[x] not in delimiters:                # move right to closest delimiter
                for x in range(x, len(l)):
                    if l[x] in delimiters:
                        cursor.x = x
                        return
                cursor.x = len(l)
            else:                                       # move right to closest word
                for x in range(x, len(l)):
                    if l[x] not in delimiters:
                        cursor.x = x
                        return
                cursor.x = len(l)

        if key in self.shortcuts['move_to_start_of_word']:
            if x == 0 and y == 0:
                pass

            elif x == 0 and y > 0:                        # move to end of line above
                cursor.y -= 1
                cursor.x = len(lines[y-1])

            elif l[x-1] not in delimiters:              # move left to closest delimiter
                for x in range(x-1, 0, -1):
                    if l[x] in delimiters:
                        cursor.x = x+1
                        return
                cursor.x = 0
            else:                                       # move left to closest word
                for x in range(x-1, 0, -1):
                    if l[x] not in delimiters:
                        cursor.x = x+1
                        return
                cursor.x = 0

        if key in self.shortcuts['select_word_left']:
            self.selection[1] = self.cursor.position
            self.input(self.shortcuts['move_to_start_of_word'][0])
            # self.selection[0] = self.cursor.position
            print(self.selection)
            self.draw_selection()


        k = key.replace(' hold', '')
        k = k.replace('shift+', '')
        # print(Text.get_width('f'))
        if len(k) == 1 and not held_keys['control']:

            if held_keys['shift']:
                k = k.upper()
                if k in self.shifted_keys:
                    k = self.shifted_keys[k]
                if k in self.alted_keys:
                    k = self.alted_keys[k]

            if self.selection:  # delete selected text
                self.delete_selected()

            add_text(k)

            if k == '(':
                add_text(')')
                cursor.x -= 1

            if k == '[':
                add_text(']')
            if k == '{':
                add_text('}')


        if key in ('space', 'space hold'):
            add_text(' ')

        if key in self.shortcuts['newline'] and self.cursor.y < self.max_lines-1:
            if l.startswith('class ') and not l.endswith(':'):
                add_text(':')
            if l.startswith('def ') and not l.endswith(':'):
                add_text(':')

            add_text('\n')

            if l.strip() == '':
                indent = len(l)
            else:
                indent = len(l) - len(l.lstrip())
                if l.lstrip().startswith('class ') or l.lstrip().startswith('def '):
                    indent += 4

            self.cursor.y += 1
            self.cursor.x = 0
            add_text(' '*indent)
            # self.cursor.x = indent

        if key in self.shortcuts['indent']:
            if self.selection == None or self.selection[0] == self.selection[1]:
                add_text(' '*4, move_cursor=True)

            else:
                for y in range(self.selection[0][1], self.selection[1][1]+1):
                    # print('indent', y)
                    lines[y] = (' '*4) + lines[y]

                self.cursor.x += 4
                self.on_undo.append((self.text, y, x))
                self.text = '\n'.join(lines)


        if key in self.shortcuts['dedent']:
            if self.selection == None:
                if l.startswith(' '*4):
                    lines[y] = l[4:]
            else:
                for y in range(self.selection[0][1], self.selection[1][1]+1):
                    l = lines[y]
                    if l.startswith(' '*4):
                        # print('dedent')
                        lines[y] = l[4:]

            self.cursor.x -= 4
            self.cursor.x = max(self.cursor.x, 0)
            self.on_undo.append((self.text, y, x))
            self.text = '\n'.join(lines)


        if key in self.shortcuts['erase']:
            if not self.selection:
                erase()
            else:
                self.delete_selected()

        if key in self.shortcuts['erase_word']:
            # print('delete word')
            if x == 0:
                erase()
                return

            if not ' ' in l:
                l = l[x:]
                self.cursor.x = 0

            else:
                removed = ''
                beginning = l[:x]


                if beginning.endswith('  '): # delete whitespace
                    beginning = beginning.rstrip()
                    removed = ' ' * (x-len(beginning))

                for delimiter in ('.', '\'', '\"', '(', '"', '\''):
                    beginning = beginning.replace(delimiter, ' ')

                if not ' ' in beginning: # first word of the line
                    beginning = ''
                else:
                    beginning = beginning.rstrip().rsplit(' ', 1)[0]

                removed = l[len(beginning):x]

                l = beginning + l[x:]
                self.cursor.x -= len(removed)


            lines[y] = l
            self.on_undo.append((self.text, y, x))
            self.text = '\n'.join(lines)


        if key in self.shortcuts['move_line_down'] and self.cursor.y < self.max_lines:
            # print('move down')
            if y+1 == len(lines): # if at last line
                self.text += '\n'

            self.move_line(y, y+1)
            cursor.y += 1

        if key in self.shortcuts['move_line_up'] and y > 0:
            self.move_line(y, y-1)
            cursor.y -= 1


        if key in self.shortcuts['undo']:
            if not on_undo:
                return

            # self.on_redo.append((self.text, y, x))
            self.text = on_undo[-1][0]
            cursor.y = on_undo[-1][1]
            cursor.x = on_undo[-1][2]
            on_undo.pop()

        # if key in self.shortcuts['redo']:
        #     if not self.on_redo:
        #         return
        #     self.text = self.on_redo[-1][0]
        #     cursor.y = self.on_redo[-1][1]
        #     cursor.x = self.on_redo[-1][2]
        #     self.on_redo.pop()


        if key in self.shortcuts['delete_line']:
            lines.pop(y)

            if y == 0:
                self.cursor.x = 0
            else:
                self.cursor.x = len(lines[y-1])

            if y >= len(lines) and y > 0:
                self.cursor.y -= 1

            self.text = '\n'.join(lines)


        if key == self.shortcuts['select_word']:
            pass
            # move cursor to the begining
            # select word right

        if key in self.shortcuts['copy']:
            print('-----copy:', self.selection)

        if key in self.shortcuts['paste']:
            self.add_text(pyperclip.paste())


        if self.register_mouse_input:
            if key == 'left mouse down' and mouse.hovered_entity == self.bg:
                # if mouse.x < self.x:
                #     return
                # from math import floor
                x = round(mouse.point[0] * self.bg.scale_x)
                y = floor(mouse.point[1] * self.bg.scale_y)

                cursor.position = (x, y)
                click_position = cursor.position
                self.selection = [click_position, click_position]


            if key == 'left mouse up':
                # if mouse.x < self.x:
                #     return

                x = round(mouse.point[0] * self.bg.scale_x)
                y = floor(mouse.point[1] * self.bg.scale_y)

                cursor.position = (x, y)
                click_position = cursor.position
                self.selection[1] = click_position

                if self.selection[1][1] < self.selection[0][1]:
                    self.selection = [self.selection[1], self.selection[0]]

                self.draw_selection()

        self.render()


    def render(self):
        lines = self.text.split('\n')
        text = '\n'.join(lines[0:self.max_lines+1])

        if self.replacements:
            self.text_entity.text = multireplace(text, self.replacements)
        else:
            self.text_entity.text = text
        self.line_numbers.text = '\n'.join([str(e) for e in range(min(len(lines), self.max_lines))])
        self.line_numbers.color = color.gray



    def update(self):
        # self.debug_cursor.position = self.get_mouse_position()

        if self.register_mouse_input and mouse.left:
            x = round(mouse.point[0] * self.bg.scale_x)
            y = floor(mouse.point[1] * self.bg.scale_y)

            self.cursor.position = (x, y)
            click_position = self.cursor.position
            self.selection[1] = click_position

            # if self.selection[1][1] < self.selection[0][1]:
            #     self.selection = [self.selection[1], self.selection[0]]

            self.draw_selection()


    def select_all(self):
        lines = self.text.split('\n')
        lines = self.text.splitlines()
        # print('|||||||||', len(lines), len(self.text.splitlines()), self.text.splitlines(), lines)
        if lines:
            self.selection = [(0,0), (len(lines[-1]), len(lines))]
            print(self.selection)


    def draw_selection(self):
        [destroy(c) for c in self.selection_parent.children]

        if self.selection == None:
            return


        if self.selection[0] == self.selection[1]:
            return

        start_y = int(self.selection[0][1])
        end_y = int(self.selection[1][1])
        lines = self.text.split('\n')

        # draw selection
        for y in range(int(self.selection[0][1]), int(self.selection[1][1])):
            e = Entity(parent=self.selection_parent, model='cube', origin=(-.5, -.5),
                color=color.color(120,1,1,.1), double_sided=True, position=(0,y*1.15), ignore=True, scale_x=len(lines[y]))
            if y == self.selection[0][1]:
                e.x = self.selection[0][0]
                e.scale_x -= self.selection[0][0]

        e = Entity(parent=self.selection_parent, model='cube', origin=(-.5, -.5),
            color=color.color(120,1,1,.1), double_sided=True, position=(0,end_y*1.15),
            ignore=True, scale_x=self.selection[1][0])
        if self.selection[0][1] == self.selection[1][1]:
            e.x = self.selection[0][0]
            e.scale_x -= self.selection[0][0]


if __name__ == '__main__':
    app = Ursina()
    # camera.orthographic = True
    # camera.fov = 1
    # window.size = window.fullscreen_size
    window.x = 200

    window.color = color.color(0, 0, .1)
    Button.color = color._20
    window.color = color._25

    # Text.size = 1/window.fullscreen_size[1]*16
    Text.default_font = 'consola.ttf'
    Text.default_resolution = 16*2
    # TreeView()
    te = TextField(max_lines=300, scale=1)
    # te.line_numbers.enabled = True
    # for name in color.color_names:
    #     if name == 'black':
    #         continue
    #     te.replacements[f' {name}'] = f'☾{name}☽ {name}☾default☽'
    # te.replacements = {
    #     'class ':    '☾orange☽class ☾default☽',
    #     'def ':      '☾azure☽def ☾default☽',
    #     '__init__':  '☾cyan☽__init__☾default☽',
    #     'Entity':    '☾lime☽Entity☾default☽',
    #     'self.':     '☾orange☽self☾default☽.',
    #     '(self)':     '(☾orange☽self☾default☽)',
    #     'self,':     '☾orange☽self☾default☽,',
    #     '    ':      '☾dark_gray☽----☾default☽',
    #     }
    #
    te.text = dedent('''
        Lorem ipsum dolor sit amet, consectetur adipiscing elit.
        Aliquam sapien tellus, venenatis sit amet ante et, malesuada porta risus.
        Etiam et mi luctus, viverra urna at, maximus eros. Sed dictum faucibus purus,
        nec rutrum ipsum condimentum in. Mauris iaculis arcu nec justo rutrum euismod.
        Suspendisse dolor tortor, congue id erat sit amet, sollicitudin facilisis velit.'''
        )[1:]
    # te.cursor.position = (4,0)
    # te.selection = [(25,0), (10,3)]
    # te.draw_selection()
    te.render()
    # te.selection = ((0,0),(4,0))
    # te.select_all()
    # te.selection = ((1,0),(3,0))
    # te.selection = ((2,3),(0,0))
    # te.draw_selection()

    app.run()



from ursina import *

class ExitButton(Button):
    def __init__(self, **kwargs):
        super().__init__(
            name = 'exit_button',
            eternal = True,
            origin = (.5, .5),
            # text_origin = (-.5,-.5),
            position = window.top_right,
            z = -999,
            scale = (.05, .025),
            color = color.red.tint(-.2),
            text = 'x',
            **kwargs)


    def on_click(self):
        application.quit()


    def input(self, key):
        if held_keys['shift'] and key == 'q' and not mouse.right:
            self.on_click()


if __name__ == '__main__':

    '''
    This is the button in the upper right corner.
    You can click on it or press Shift+Q to close the program.
    To disable it, set window.exit_button.enabled to False
    '''
    app = Ursina()
    app.run()



from ursina import *


class CollisionZone(Entity):
    def __init__(self, radius=2, target_entities=None, **kwargs):
        super().__init__(**kwargs)
        self.radius = radius
        self.entities_with_mesh_colliders = target_entities # defaults to all entities with a mesh collider
        if not self.entities_with_mesh_colliders:
            self.entities_with_mesh_colliders = [e for e in scene.entities if isinstance(e.collider, MeshCollider)]

        self._update_rate = .25
        self._t = self._update_rate
        self._prev_pos = self.world_position
        # self.debug = False
        # self.debug_parent = Entity(ignore=True)
        for e in self.entities_with_mesh_colliders:
            e.collider.node.clearSolids()
        self.update_colliders()


    def update(self):
        self._t += time.dt
        if self._t < self._update_rate:
            return

        self._t = 0

        # only update after we've moved a certain distance
        if distance_xz(self.world_position, self._prev_pos) < self.radius / 2:
            return

        self._prev_pos = self.world_position
        self.update_colliders()


    def update_colliders(self):
        # if self.debug:
        #     destroy(self.debug_parent)
        # self.debug_parent = Entity(ignore=True)


        for e in self.entities_with_mesh_colliders:
            # if self.debug:
            #     # print('debug')
            #     debug_model = Entity(
            #         parent=self.debug_parent,
            #         world_position=e.world_position,
            #         world_scale=e.world_scale,
            #         model=Mesh(mode='line', static=False),
            #         color=color.red,
            #         always_on_top=True
            #     ).model
            #     for tri in e.collider.collision_polygons:
            #         if (distance_xz(self.get_position(e), tri.getPoint(0)) < self.radius / e.world_scale_x
            #         or distance_xz(self.get_position(e), tri.getPoint(1)) < self.radius / e.world_scale_x
            #         or distance_xz(self.get_position(e), tri.getPoint(2)) < self.radius / e.world_scale_x):
            #             if self.debug:
            #                 debug_model.vertices.extend([tri.getPoint(2), tri.getPoint(1), tri.getPoint(0)])
            #
            #     # debug_model.generate()


            e.collider.node.clearSolids()
            i = 0
            for tri in e.collider.collision_polygons:
                if (distance_xz(self.get_position(e), tri.getPoint(0)) < self.radius / e.world_scale_x
                or distance_xz(self.get_position(e), tri.getPoint(1)) < self.radius / e.world_scale_x
                or distance_xz(self.get_position(e), tri.getPoint(2)) < self.radius / e.world_scale_x):
                    # e.collider.node.addSolid(tri)
                    invoke(e.collider.node.addSolid, tri, delay=i*.01)
                    i+= 1


if __name__ == '__main__':
    '''
    This will only enable mesh colliders' collision polygons within a certain range,
    in order to improve performance.
    '''

    from ursina.shaders import basic_lighting_shader
    window.vsync = False
    app = Ursina()

    application.asset_folder = application.asset_folder.parent
    terrain = Entity(model='heightmap_test', scale=32, texture='grass', collider='mesh', shader=basic_lighting_shader)
    from ursina.prefabs.first_person_controller import FirstPersonController
    player = FirstPersonController(speed=10)
    collision_zone = CollisionZone(parent=player, radius=32)


    def input(key):
        if key == 'c':
            terrain.collision = not terrain.collision


    # def update():
    #     collision_zone.x += (held_keys['d'] - held_keys['a']) * time.dt * 2
    #     collision_zone.z += (held_keys['w'] - held_keys['s']) * time.dt * 2
    #
    #     hit_info = raycast(collision_zone.world_position+Vec3(0,2,0), Vec3(0,-1,0))
    #     if hit_info.hit:
    #         collision_zone.y = hit_info.world_point.y

    Sky()
    # EditorCamera(position=(0,0,0))
    base.set_frame_rate_meter(True)
    app.run()



from ursina import *


class Cylinder(Prismatoid):
    def __init__(self, resolution=8, radius=.5, start=0, height=1, direction=(0,1,0), mode='triangle', **kwargs):
        super().__init__(
            base_shape=Circle(resolution=resolution, radius=.5),
            origin=(0,0),
            path=((0,start,0), Vec3(direction) * (height+start)),
            thicknesses=((radius*2, radius*2),),
            mode=mode,
            **kwargs
            )


if __name__ == '__main__':
    app = Ursina()
    Entity(model=Cylinder(6, start=-.5), color=color.color(60,1,1,.3))
    origin = Entity(model='quad', color=color.orange, scale=(5, .05))
    ed = EditorCamera(rotation_speed = 200, panning_speed=200)
    app.run()



import time

from direct.showbase.ShowBase import ShowBase
from direct.task.Task import Task
import gltf

from ursina import application
from ursina import input_handler
from ursina.window import instance as window
from ursina.scene import instance as scene
from ursina.camera import instance as camera
from ursina.mouse import instance as mouse

import __main__
time.dt = 0


class Ursina(ShowBase):

    def __init__(self, **kwargs): # optional arguments: title, fullscreen, size, position, vsync, borderless, show_ursina_splash, render_mode, development_mode, editor_ui_enabled.
        for name in ('size', 'vsync'):
            if name in kwargs and hasattr(window, name):
                setattr(window, name, kwargs[name])

        if 'development_mode' in kwargs:
            application.development_mode = kwargs['development_mode']

        super().__init__()
        application.base = base
        gltf.patch_loader(self.loader)

        window.late_init()
        for name in ('title', 'fullscreen', 'position', 'show_ursina_splash', 'borderless', 'render_mode'):
            if name in kwargs and hasattr(window, name):
                setattr(window, name, kwargs[name])

        # camera
        camera._cam = base.camera
        camera._cam.reparent_to(camera)
        camera.render = base.render
        camera.position = (0, 0, -20)
        scene.camera = camera
        camera.set_up()

        # input
        base.buttonThrowers[0].node().setButtonDownEvent('buttonDown')
        base.buttonThrowers[0].node().setButtonUpEvent('buttonUp')
        base.buttonThrowers[0].node().setButtonRepeatEvent('buttonHold')
        self._input_name_changes = {
            'mouse1' : 'left mouse down',
            'mouse1 up' : 'left mouse up',
            'mouse2' : 'middle mouse down',
            'mouse2 up' : 'middle mouse up',
            'mouse3' : 'right mouse down',
            'mouse3 up' : 'right mouse up',
            'wheel_up' : 'scroll up',
            'wheel_down' : 'scroll down',
            'arrow_left' : 'left arrow',
            'arrow_left up' : 'left arrow up',
            'arrow_up' : 'up arrow',
            'arrow_up up' : 'up arrow up',
            'arrow_down' : 'down arrow',
            'arrow_down up' : 'down arrow up',
            'arrow_right' : 'right arrow',
            'arrow_right up' : 'right arrow up',
            'lcontrol' : 'left control',
            'rcontrol' : 'right control',
            'lshift' : 'left shift',
            'rshift' : 'right shift',
            'lalt' : 'left alt',
            'ralt' : 'right alt',
            'lcontrol up' : 'left control up',
            'rcontrol up' : 'right control up',
            'lshift up' : 'left shift up',
            'rshift up' : 'right shift up',
            'lalt up' : 'left alt up',
            'ralt up' : 'right alt up',
            'control-mouse1' : 'left mouse down',
            'control-mouse2' : 'middle mouse down',
            'control-mouse3' : 'right mouse down',
            'shift-mouse1' : 'left mouse down',
            'shift-mouse2' : 'middle mouse down',
            'shift-mouse3' : 'right mouse down',
            'alt-mouse1' : 'left mouse down',
            'alt-mouse2' : 'middle mouse down',
            'alt-mouse3' : 'right mouse down',
            'page_down' : 'page down',
            'page_down up' : 'page down up',
            'page_up' : 'page up',
            'page_up up' : 'page up up',
            }
        self.accept('buttonDown', self.input)
        self.accept('buttonUp', self.input_up)
        self.accept('buttonHold', self.input_hold)

        base.disableMouse()
        mouse._mouse_watcher = base.mouseWatcherNode
        mouse.enabled = True
        self.mouse = mouse

        from ursina import gamepad

        scene.set_up()
        self._update_task = taskMgr.add(self._update, "update")

        # try to load settings that need to be applied before entity creation
        application.load_settings()

        from ursina import HotReloader
        # make sure it's running from a file and not an interactive session.
        application.hot_reloader = HotReloader(__main__.__file__ if hasattr(__main__, '__file__') else 'None')

        window.make_editor_gui()
        if 'editor_ui_enabled' in kwargs:
            window.editor_ui.enabled = kwargs['editor_ui_enabled']


    def _update(self, task):
        # time between frames
        dt = globalClock.getDt() * application.time_scale
        time.dt = dt

        mouse.update()

        if hasattr(__main__, 'update') and not application.paused:
            __main__.update()

        for seq in application.sequences:
            seq.update()

        for entity in scene.entities:
            if entity.enabled == False or entity.ignore:
                continue

            if application.paused and entity.ignore_paused == False:
                continue

            if hasattr(entity, 'update') and callable(entity.update):
                entity.update()


            if hasattr(entity, 'scripts'):
                for script in entity.scripts:
                    if script.enabled and hasattr(script, 'update') and callable(script.update):
                        script.update()


        return Task.cont


    def input_up(self, key):
        if key in  ('wheel_up', 'wheel_down'):
            return

        key += ' up'
        self.input(key)


    def input_hold(self, key):
        key = key.replace('control-', '')
        key = key.replace('shift-', '')
        key = key.replace('alt-', '')

        if key in self._input_name_changes:
            key = self._input_name_changes[key]

        key += ' hold'
        self.input(key)


    def input(self, key):
        key = key.replace('control-', '')
        key = key.replace('shift-', '')
        key = key.replace('alt-', '')

        if key in self._input_name_changes:
            key = self._input_name_changes[key]

        if key in input_handler.rebinds:
            key = input_handler.rebinds[key]

        try: input_handler.input(key)
        except: pass
        if not application.paused:
            if hasattr(__main__, 'input'):
                __main__.input(key)


        for entity in scene.entities:
            if entity.enabled == False or entity.ignore or entity.ignore_input:
                continue
            if application.paused and entity.ignore_paused == False:
                continue

            if hasattr(entity, 'input') and callable(entity.input):
                entity.input(key)

            if hasattr(entity, 'scripts'):
                for script in entity.scripts:
                    if script.enabled and hasattr(script, 'input') and callable(script.input):
                        script.input(key)


        try: mouse.input(key)
        except: pass


        if key == 'f12':
            window.editor_ui.enabled = not window.editor_ui.enabled

        if key == 'f11':
            window.fullscreen = not window.fullscreen

        if key == 'f10':
            i = window.render_modes.index(window.render_mode)
            i += 1
            if i >= len(window.render_modes):
                i = 0

            window.render_mode = window.render_modes[i]

        if key == 'f9':
            window.render_mode = 'default'


    def run(self):
        if window.show_ursina_splash:
            from ursina.prefabs import ursina_splash

        application.load_settings()
        print('development mode:', application.development_mode)
        print('application successfully started')

        super().run()


if __name__ == '__main__':
    app = Ursina()
    app.run()



import operator
from math import sqrt, sin, acos
from panda3d.core import Vec4, LVector3f, Mat3, Mat4
from ursina.vec2 import Vec2
from ursina.vec3 import Vec3
from ursina.color import Color
internal_sum = sum


def distance(a, b):
    try:
        # dist = [abs(e) for e in (a - b)]
        dist = abs(a[0] - b[0])
        dist += abs(a[1] - b[1])
        dist += abs(a[2] - b[2])
        dist += abs(a[3] - b[3])
        # print('color distance', a, b)
        return dist
    except:
        pass

    # if input is Entity, convert to positions
    if hasattr(a, 'position'): a = a.position
    if hasattr(b, 'position'): b = b.position

    dist = sqrt((b[0] - a[0])**2 + (b[1] - a[1])**2 + (b[2] - a[2])**2)
    # print('------------DIST:', dist)
    return dist


def distance2d(a, b):
    if hasattr(a, 'position'): a = a.position
    if hasattr(b, 'position'): b = b.position

    return sqrt((b[0] - a[0])**2 + (b[1] - a[1])**2)


def distance_xz(a, b):
    if hasattr(a, 'position'): a = a.position
    if hasattr(b, 'position'): b = b.position

    return sqrt((b[0] - a[0])**2 + (b[2] - a[2])**2)


def lerp(a, b, t):
    if isinstance(a, (int, float, complex)):
        return a + (b - a) * t

    elif isinstance(a, Color) and isinstance(b, Color):
        col = [lerp(e[0], e[1], t) for e in zip(a,b)]
        return Color(col[0], col[1], col[2], col[3])

    elif isinstance(a, (tuple, list, Vec2, Vec3, Vec4, LVector3f)) and isinstance(b, (tuple, list, Vec2, Vec3, Vec4, LVector3f)):
        lerped = list()
        for i in range(min(len(a), len(b))):
            lerped.append(lerp(a[i], b[i], t))

        if isinstance(a, (tuple, list)):
            return type(a)(lerped)
        else:
            return type(a)(*lerped)
    else:
        print(f'''can't lerp types {type(a)} and {type(b)}''')


def inverselerp(a, b, t) :
    return (a - b) / (t - b)


def slerp(q1, q2, t):
    costheta = q1.dot(q2)
    if costheta < 0.0:
        costheta = -costheta
        q1 = q1.conjugate()
    elif costheta > 1.0:
        costheta = 1.0

    theta = acos(costheta)
    if abs(theta) < 0.01:
        return q2

    sintheta = sqrt(1.0 - costheta * costheta)
    if abs(sintheta) < 0.01:
        return (q1+q2)*0.5

    r1 = sin((1.0 - t) * theta) / sintheta
    r2 = sin(t * theta) / sintheta
    return (q1*r1) + (q2*r2)


def clamp(value, floor, ceiling):
    return max(min(value, ceiling), floor)


def round_to_closest(value, step=0):
    if not step:
        return value

    step = 1/step
    return round(value * step) / step


def chunk_list(l, chunk_size):
    # yield successive chunks from list
    for i in range(0, len(l), chunk_size):
        yield l[i:i + chunk_size]


def size_list():
    #return a list of current python objects sorted by size
    globals_list = list()
    globals_list.clear()
    for e in globals():
        # object, size
        globals_list.append([e, sys.getsizeof(e)])
    globals_list.sort(key=operator.itemgetter(1), reverse=True)
    print('scene size:', globals_list)


def sum(l):
    try:
        return internal_sum(l)
    except:
        pass

    total = l[0].__class__()
    for e in l:
        total += e

    return total


if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    e1 = Entity(position = (0,0,0))
    e2 = Entity(position = (0,1,1))
    distance(e1, e2)
    distance_xz(e1, e2.position)

    between_color = lerp(color.lime, color.magenta, .5)
    print(between_color)
    print(lerp((0,0), (0,1), .5))
    print(lerp(Vec2(0,0), Vec2(0,1), .5))
    print(lerp([0,0], [0,1], .5))

    print(round(Vec3(.38, .1351, 353.26), 2))

    app.run()



from ursina import *

if __name__ == '__main__':
    app = Ursina()

camera.overlay.color = color.black
logo = Sprite(name='ursina_splash', parent=camera.ui, texture='ursina_logo', world_z=camera.overlay.z-1, scale=.1, color=color.clear)
logo.animate_color(color.white, duration=2, delay=1, curve=curve.out_quint_boomerang)
camera.overlay.animate_color(color.clear, duration=1, delay=4)
destroy(logo, delay=5)

def splash_input(key):
    destroy(logo)
    camera.overlay.animate_color(color.clear, duration=.25)

logo.input = splash_input



if __name__ == '__main__':
    app.run()



from ursina import *
import pyperclip
from copy import deepcopy
import sys
from math import floor



class GridEditor(Entity):
    def __init__(self, size=(32,32), palette=(' ', '#', '|', 'o'), **kwargs):
        super().__init__(parent=camera.ui, model='quad', position=(-.45,-.45), origin=(-.5,-.5), scale=.9, collider='box')

        self.w, self.h = int(size[0]), int(size[1])
        sys.setrecursionlimit(self.w * self.h)
        # self.grid = [[palette[0] for x in range(self.w)] for y in range(self.h)]
        if not hasattr(self, 'grid'):
            self.grid = [[palette[0] for y in range(self.h)] for x in range(self.w)]
        self.brush_size = 1
        self.auto_render = True
        self.cursor = Entity(parent=self, model=Quad(segments=0, mode='line'), origin=(-.5,-.5), scale=(1/self.w, 1/self.h), color=color.color(0,1,1,.5), z=-.1)

        self.selected_char = palette[1]
        self.palette = palette
        self.prev_draw = None
        self.start_pos = (0,0)
        self.outline = Entity(parent=self, model=Quad(segments=0, mode='line', thickness=1), color=color.cyan, z=.01, origin=(-.5,-.5))

        self.undo_cache = list()
        self.undo_cache.append(deepcopy(self.grid))
        self.undo_index = 0

        self.help_text = Text(
            text=dedent('''
                left mouse:    draw
                control(hold): draw lines
                alt(hold):     select character
                ctrl + z:      undo
                ctrl + y:      redo
            '''),
            position=window.top_left,
            scale=.75
            )

        self.edit_mode = True

        for key, value in kwargs.items():
            setattr(self, key ,value)



    @property
    def palette(self):
        return self._palette

    @palette.setter
    def palette(self, value):
        self._palette = value
        if hasattr(self, 'palette_parent'):
            destroy(self.palette_parent)

        self.palette_parent = Entity(parent=camera.ui, position=(-.75,-.05))
        for i, e in enumerate(value):
            if isinstance(e, str):
                i = e

            b = Button(parent=self.palette_parent, scale=.05, text=i, model='quad', color=color._32)
            b.on_click = Func(setattr, self, 'selected_char', e)
            b.tooltip = Tooltip(str(e))

            if isinstance(e, Color):
                b.color = e

        grid_layout(self.palette_parent.children, max_x=4)


    @property
    def edit_mode(self):
        return self._edit_mode

    @edit_mode.setter
    def edit_mode(self, value):
        self._edit_mode = value
        self.cursor.enabled = value
        self.outline.enabled = value
        self.palette_parent.enabled = value
        self.help_text.enabled = value


    def update(self):
        if not self.edit_mode:
            return

        self.cursor.enabled = mouse.hovered_entity == self
        if self.hovered:
            self.cursor.position = mouse.point
            self.cursor.x = floor(self.cursor.x * self.w) / self.w
            self.cursor.y = floor(self.cursor.y * self.h) / self.h

            if mouse.left:
                y = int(round(self.cursor.y * self.h))
                x = int(round(self.cursor.x * self.w))
                # if (x, y) == self.prev_draw:
                #     return

                if not held_keys['alt']:
                    if self.prev_draw is not None and distance2d(self.prev_draw, (x,y)) > 1:
                        dist = distance2d(self.prev_draw, (x,y))

                        if dist > 1: # draw line
                            for i in range(int(dist)+1):
                                inbetween_pos = lerp(self.prev_draw, (x,y), i/dist)
                                self.draw(int(inbetween_pos[0]), int(inbetween_pos[1]))

                            self.draw(x, y)
                            self.prev_draw = (x,y)

                    else:
                        self.draw(x, y)
                        self.prev_draw = (x,y)

                else:
                    self.selected_char = self.grid[x][y]



    def draw(self, x, y):
        for _y in range(y, min(y+self.brush_size, self.h)):
            for _x in range(x, min(x+self.brush_size, self.w)):
                self.grid[_x][_y] = self.selected_char

        if self.auto_render:
            self.render()


    def input(self, key):
        if key == 'tab':
            self.edit_mode = not self.edit_mode


        # if not self.edit_mode:
        #     return

        if key == 'left mouse down':
            self.start_pos = (
                int(self.cursor.x * (self.w+1)),
                -int(self.cursor.y * self.h)
            )
            if not held_keys['shift']:
                self.prev_draw = None

        if key == 'left mouse up':
            self.render()

            if not held_keys['control']:
                self.record_undo()


        if held_keys['control'] and key == 'z':
            self.undo_index -= 1
            self.undo_index = clamp(self.undo_index, 0, len(self.undo_cache)-1)
            self.grid = deepcopy(self.undo_cache[self.undo_index])
            self.render()

        if held_keys['control'] and key == 'y':
            self.undo_index += 1
            self.undo_index = clamp(self.undo_index, 0, len(self.undo_cache)-1)
            self.grid = deepcopy(self.undo_cache[self.undo_index])
            self.render()

        # fill
        if key == 'g':
            y = int(self.cursor.y * self.h)
            x = int(self.cursor.x * self.w)
            self.floodfill(self.grid, x, y)
            self.render()
            self.record_undo()

        if key == 'x' and self.brush_size > 1:
            self.brush_size -= 1
            self.cursor.scale = Vec2(self.brush_size / self.w, self.brush_size / self.h)
            self.prev_draw = None

        if key == 'd' and self.brush_size <  8:
            self.brush_size += 1
            self.cursor.scale = Vec2(self.brush_size / self.w, self.brush_size / self.h)
            self.prev_draw = None

        if held_keys['control'] and key == 's':
            if hasattr(self, 'save'):
                self.save()


    def record_undo(self):
        self.undo_index += 1
        self.undo_cache = self.undo_cache[:self.undo_index]
        self.undo_cache.append(deepcopy(self.grid))



    def floodfill(self, matrix, x, y, first=True):
        if matrix[x][y] == self.selected_char:
            return

        if first:
            self.fill_target = matrix[x][y]

        if matrix[x][y] == self.fill_target:
            matrix[x][y] = self.selected_char
            # recursively invoke flood fill on all surrounding cells
            if x > 0:
                self.floodfill(matrix, x-1, y, first=False)
            if x < self.w-1:
                self.floodfill(matrix, x+1, y, first=False)
            if y > 0:
                self.floodfill(matrix, x, y-1, first=False)
            if y < self.h-1:
                self.floodfill(matrix, x, y+1, first=False)




class PixelEditor(GridEditor):
    def __init__(self, texture, palette=(color.black, color.white, color.light_gray, color.gray, color.red, color.orange, color.yellow, color.lime, color.green, color.turquoise, color.cyan, color.azure, color.blue, color.violet, color.magenta, color.pink), **kwargs):
        self.grid = [[texture.get_pixel(x,y) for y in range(texture.height)] for x in range(texture.width)]
        super().__init__(texture=texture, size=texture.size, palette=palette, **kwargs)
        self.texture.filtering = False
        self.render()


    def render(self):
        for y in range(self.h):
            for x in range(self.w):
                self.texture.set_pixel(x, y, self.grid[x][y])

        self.texture.apply()


    def save(self):
        if self.texture.path:
            self.texture.save(self.texture.path)
            print('saving:', self.texture.path)



class ASCIIEditor(GridEditor):
    def __init__(self, size=(61,28), palette=(' ', '#', '|', 'A', '/', '\\', 'o', '_', '-', 'i', 'M', '.'), font='VeraMono.ttf', color=color.black, line_height=1.1, **kwargs):
        super().__init__(size=size, palette=palette, color=color, **kwargs)
        rotated_grid = list(zip(*self.grid[::-1]))
        text = '\n'.join([''.join(reversed(line)) for line in reversed(rotated_grid)])

        self.text_entity = Text(parent=self.parent, text=text, x=self.x, line_height=line_height, font=font)

        self.scale = (self.text_entity.width, self.text_entity.height)
        self.text_entity.world_parent = self
        self.text_entity.y = 1
        self.text_entity.z = -.001

    def render(self):
        rotated_grid = list(zip(*self.grid[::-1]))
        self.text_entity.text = '\n'.join([''.join(reversed(line)) for line in reversed(rotated_grid)])


    def input(self, key):
        super().input(key)
        if held_keys['control'] and key == 'c':
            print(self.text_entity.text)
            pyperclip.copy(self.text_entity.text)
        #
        # if held_keys['control'] and key == 'v' and pyperclip.paste().count('\n') == (h-1):
        #     t.text = pyperclip.paste()
        #     undo_index += 1
        #     undo_cache = undo_cache[:undo_index]
        #     undo_cache.append(deepcopy(grid))




if __name__ == '__main__':
    app = Ursina()


    '''
    pixel editor example, it's basically a drawing tool.
    can be useful for level editors and such
    here we create a new texture, but can also give it an exisitng texture to modify.
    '''
    from PIL import Image
    t = Texture(Image.new(mode='RGBA', size=(32,32), color=(0,0,0,1)))
    from ursina.prefabs.grid_editor import PixelEditor
    PixelEditor(texture=load_texture('brick'))

    '''
    same as the pixel editor, but with text.
    '''
    from ursina.prefabs.grid_editor import ASCIIEditor
    ASCIIEditor()


    app.run()



from ursina import *
from ursina import distance as ursina_distance
from math import inf
from ursina.hit_info import HitInfo



def prepare_terrain(terrain, debug=False, calculate_normals=True):

    #does calculations that are needed for every terraincast check, and sets up the entities that manage transofrmations
    #has the option to calculate all face normals for the terrain (quicker while running, but takes time to initiate)

    terrain._cast = Entity(parent=terrain,
                           model='sphere',
                           color=color.orange,
                           world_scale=0.1,
                           position=(0, 1, 0),
                           visible=debug,)
    if debug:
        terrain._cast.bound = Entity(parent=terrain,
                                     model='cube',
                                     color=color.rgba(255, 0, 0, 100))

    terrain._cast.direction = Entity(parent=terrain,
                                     position=(0, 1, 0))

    height_values = terrain.model.height_values

    terrain._cast.width = terrain.model.width
    terrain._cast.depth = terrain.model.depth
    terrain._cast.aspect_ratio = terrain._cast.depth / terrain._cast.width

    #correction_scale needed to account for cases where the other dimension is the smaller, so treated as 1 now
    if terrain._cast.depth < terrain._cast.width:
        terrain._cast.correction_scale = 1 / terrain._cast.aspect_ratio
    else:
        terrain._cast.correction_scale = 1

    terrain._cast.max = max([max(i) for i in height_values])
    terrain._cast.min = min([min(i) for i in height_values])

    if calculate_normals:
        terrain._cast.prepared_height_values = []
        for scan_x, v in enumerate(height_values):
            row_to_add = []
            for scan_z, w in enumerate(v):
                quad_to_add = []
                for sub_face in [False, True]:
                    quad_to_add.append(_terraincast_get_plane(terrain, scan_x, scan_z, sub_face))
                row_to_add.append(quad_to_add)
            terrain._cast.prepared_height_values.append(row_to_add)


def _terraincast_get_plane(terrain, scan_x, scan_z, sub_face):
    #gets details needed for each plane that makes up the terrain
    from numpy import cross
    height_values = terrain.model.height_values
    if scan_z == len(height_values[0]) - 1 and scan_x == len(height_values) - 1:
        start = Vec3(scan_x, height_values[scan_x][scan_z], scan_z)
        right = Vec3(scan_x + 1, height_values[scan_x][scan_z], scan_z)
        left = Vec3(scan_x, height_values[scan_x][scan_z], scan_z + 1)

    elif scan_z == len(height_values[0]) - 1:
        start = Vec3(scan_x, height_values[scan_x][scan_z], scan_z)
        right = Vec3(scan_x, height_values[scan_x][scan_z], scan_z + 1)
        left = Vec3(scan_x + 1, height_values[scan_x + 1][scan_z], scan_z + 1)
        # print('edge')

    elif scan_x == len(height_values) - 1:
        start = Vec3(scan_x, height_values[scan_x][scan_z], scan_z)
        right = Vec3(scan_x, height_values[scan_x][scan_z + 1], scan_z + 1)
        left = Vec3(scan_x + 1, height_values[scan_x][scan_z + 1], scan_z + 1)

    elif sub_face:
        start = Vec3(scan_x, height_values[scan_x][scan_z], scan_z)
        right = Vec3(scan_x, height_values[scan_x][scan_z + 1], scan_z + 1)
        left = Vec3(scan_x + 1, height_values[scan_x + 1][scan_z + 1], scan_z + 1)

    else:
        start = Vec3(scan_x, height_values[scan_x][scan_z], scan_z)
        right = Vec3(scan_x + 1, height_values[scan_x + 1][scan_z], scan_z)
        left = Vec3(scan_x + 1, height_values[scan_x + 1][scan_z + 1], scan_z + 1)

    normal = cross(left - start, right - start)
    if normal[1] < 0:
        normal = - normal
    normal = normal.tolist()
    normal = Vec3(*normal).normalized()

    return start, normal


def terraincast(origin,
                terrain,
                direction=Vec3(0, -1, 0),
                distance=inf,
                iterations=inf,
                debug=False,
                calculate_normals=True,
                ):
    from numpy import dot
    origin = Vec3(*origin)
    direction = Vec3(*direction)

    height_values = terrain.model.height_values
    current_iterations = 0

    if not hasattr(terrain, "_cast"):
        prepare_terrain(terrain, debug, calculate_normals)

    #transformations to make terraincast line up with model space, then "height_values grid" space
    terrain._cast.world_position = origin
    terrain._cast.position += terrain.origin

    terrain._cast.direction.world_position = terrain.world_position + direction.normalized()

    model_origin = terrain._cast.position
    model_origin[0] = model_origin[0] / terrain._cast.correction_scale
    model_origin[2] = model_origin[2] / terrain._cast.correction_scale

    model_origin = model_origin + Vec3(0.5, 0, 0.5 * terrain._cast.aspect_ratio)

    model_direction = terrain._cast.direction.position
    model_direction[1] = model_direction[1] * terrain._cast.correction_scale

    #sets up scan information that iterates through the grid (height_values "grid" space essentially)
    scan_direction = Vec3(model_direction[0] * terrain._cast.width, model_direction[1],
                          model_direction[2] * terrain._cast.width)
    scan_tile = Vec3(model_origin[0] * terrain._cast.width, model_origin[1], model_origin[2] * terrain._cast.width)
    original_tile = Vec3(model_origin[0] * terrain._cast.width, model_origin[1], model_origin[2] * terrain._cast.width)

    scan_max_distance = ursina_distance(Vec3(0, 0, 0), scan_direction) * distance

    searching = True
    sub_face = False

    # handle case where ray starts outside model
    x_scalar = 0
    y_scalar = 0
    z_scalar = 0

    if scan_tile[0] < 0 and scan_direction[0] > 0:
        x_scalar = (0 - scan_tile[0]) / scan_direction[0]
    elif scan_tile[0] > len(height_values) and scan_direction[0] < 0:
        x_scalar = abs((scan_tile[0] - (len(height_values))) / scan_direction[0])

    if scan_tile[2] < 0 and scan_direction[2] > 0:
        z_scalar = (0 - scan_tile[2]) / scan_direction[2]
    elif scan_tile[2] > len(height_values[0]) and scan_direction[2] < 0:
        z_scalar = abs((scan_tile[2] - (len(height_values[0]))) / scan_direction[2])

    if scan_tile[1] < terrain._cast.min and scan_direction[1] > 0:
        y_scalar = (terrain._cast.min - scan_tile[1]) / scan_direction[1]
    elif scan_tile[1] > terrain._cast.max and scan_direction[1] < 0:
        y_scalar = abs((scan_tile[1] - terrain._cast.max) / scan_direction[1])

    scan_tile += scan_direction * max([x_scalar, y_scalar, z_scalar])

    #main loop through grid
    while searching:

        #determines indexes required from grid
        scan_x = int(floor(scan_tile[0]))
        scan_z = int(floor(scan_tile[2]))
        if scan_tile[0] == scan_x and scan_direction[0] < 0:
            scan_x -= 1
            scan_x = max(scan_x, 0)

        if scan_tile[2] == scan_z and scan_direction[2] < 0:
            scan_z -= 1
            scan_z = max(scan_z, 0)

        current_iterations += 1

        #failure cases
        if current_iterations == iterations:
            searching = False
        elif scan_x >= len(height_values):
            searching = False
        elif scan_z >= len(height_values[0]):
            searching = False
        elif scan_x < 0:
            searching = False
        elif scan_z < 0:
            searching = False
        elif scan_tile[1] > terrain._cast.max + 1:
            searching = False
        elif scan_tile[1] < terrain._cast.min - 1:
            searching = False
        elif ursina_distance(scan_tile, original_tile) > scan_max_distance:
            searching = False

        if not searching:
            break

        if calculate_normals:
            start, normal = terrain._cast.prepared_height_values[scan_x][scan_z][sub_face]
        else:
            start, normal = _terraincast_get_plane(terrain, scan_x, scan_z, sub_face)

        bottom_dot = dot(scan_direction, normal)
        top_dot = dot(start - original_tile, normal)
        if bottom_dot != 0:
            magnitude = top_dot / bottom_dot
        elif top_dot == 0:
            magnitude = 0
        else:
            magnitude = None

        if magnitude is not None:
            point = original_tile + magnitude * scan_direction
        else:
            magnitude = 0
            point = Vec3(scan_x + 2, 0, scan_z + 2)

        in_front = magnitude >= 0
        correct_sub_triangle = (point[0] % 1 <= point[2] % 1) == sub_face or point[0] % 1 == point[2] % 1
        x_tolerance = abs(point[0] - scan_x - 0.5) <= 0.5
        z_tolerance = abs(point[2] - scan_z - 0.5) <= 0.5
        #success case
        if in_front and correct_sub_triangle and x_tolerance and z_tolerance:
            break

        elif sub_face:
            # works out next "tile" to move to
            if scan_direction[0] > 0:
                x_scalar = (1 - scan_tile[0] % 1) / scan_direction[0]
            elif scan_direction[0] < 0:
                corrected = (-scan_tile[0]) % 1
                x_scalar = abs((1 - corrected) / scan_direction[0])
            else:
                x_scalar = None

            if scan_direction[2] > 0:
                z_scalar = (1 - scan_tile[2] % 1) / scan_direction[2]
            elif scan_direction[2] < 0:
                corrected = (-scan_tile[2]) % 1
                z_scalar = abs((1 - corrected) / scan_direction[2])
            else:
                z_scalar = None

            if x_scalar is not None and (z_scalar is None or x_scalar < z_scalar):
                scan_tile += x_scalar * scan_direction
            elif z_scalar is not None:
                scan_tile += z_scalar * scan_direction
            else:
                searching = False
                break

        sub_face = not sub_face

    if debug:
        #generates bounding box
        l = -0.5 * terrain._cast.correction_scale - terrain.origin[0]
        r = 0.5 * terrain._cast.correction_scale - terrain.origin[0]

        f = 0.5 * terrain._cast.aspect_ratio * terrain._cast.correction_scale - terrain.origin[2]
        b = -0.5 * terrain._cast.aspect_ratio * terrain._cast.correction_scale - terrain.origin[2]

        u = terrain._cast.max - terrain.origin[1]
        d = terrain._cast.min - terrain.origin[1]

        verts = (
            Vec3(l, d, b), Vec3(r, d, b), Vec3(r, u, b), Vec3(l, u, b),
            Vec3(l, d, f), Vec3(r, d, f), Vec3(r, u, f), Vec3(l, u, f)
        )

        tris = (
            (0, 1, 2, 3), (5, 4, 7, 6),  # forward, back
            (3, 2, 6, 7), (4, 5, 1, 0),  # up, down
            (1, 5, 6, 2), (4, 0, 3, 7)  # right, left
        )
        cube = Mesh(verts, tris)
        terrain._cast.bound.model = cube
        terrain._cast.bound.model.generate()
    if searching:
        #geneartess hit info and undoes some transformations
        point = Vec3(point[0] / terrain._cast.width,
                     point[1],
                     point[2] / terrain._cast.width)

        point -= Vec3(0.5, 0, 0.5 * terrain._cast.aspect_ratio)

        point[0] = point[0] * terrain._cast.correction_scale
        point[2] = point[2] * terrain._cast.correction_scale
        terrain._cast.position = point
        terrain._cast.position -= terrain.origin

        terrain._cast.visible = debug

        terrain._cast.direction.position = normal / terrain._cast.correction_scale

        hit = HitInfo(hit=True)
        hit.point = point
        hit.world_point = terrain._cast.world_position

        hit.normal = Vec3(*normal)
        hit.world_normal = terrain._cast.direction.world_position - terrain.world_position
        hit.distance = ursina_distance(origin, hit.world_point)
        hit.entity = terrain
        hit.entities = [terrain, ]
        hit.hits = [True, ]
        if hit.distance > distance:
            hit = HitInfo(hit=False)
        return hit
    else:
        terrain._cast.visible = False

        hit = HitInfo(hit=False)
        return hit


if __name__ == '__main__':
    app = Ursina()

    terrainEntity = Entity(model=Terrain('heightmap_1', skip=8),
                           scale=(20, 5, 20),
                           rotation=(30, 40, 50),
                           origin=(1, 1, 1),
                           texture='heightmap_1')

    hit_entity = Entity(model='sphere', scale=0.1)
    EditorCamera()


    def update():
        hit = terraincast(camera.world_position, terrainEntity, direction=camera.forward, debug=True)
        if hit:
            hit_entity.position = hit.world_point + hit.world_normal

            hit.entity.rotation_y += 2*time.dt

    Sky()
    app.run()



from ursina import *


def grid_layout(l, max_x=8, max_y=8, spacing=(0,0,0), origin=(-.5,.5,0), offset=(0,0,0)):
    if len(origin) == 2:
        origin += (0,)
    if len(offset) == 2:
        offset += (0,)
    if not isinstance(l, list):
        print('error: grid_layout input must be a list or tuple, not', l.__class__.__name__)
        return
    x, y, z = 0, 0, 0

    dimensions = l[0].bounds
    direction = [-e*2 for e in origin]
    direction = [1 if e == 0 else e for e in direction]
    # print(direction)
    row = list()

    for i, e in enumerate(l):
        e.position = (
            x * dimensions[0] * direction[0],
            y * dimensions[1] * direction[1],
            z * dimensions[2] * direction[2]
        )
        e.position += offset
        e.origin = origin
        row.append(e)

        x += 1
        if x >= max_x:
            y += 1
            x = 0
            # center row
            if origin[0] == 0:
                for e in row:
                    e.x -= e.bounds[0] * len(row) / 2 - e.bounds[0] / 2
            row.clear()

        if y >= max_y:
            z += 1
            y = 0


    # center last row
    if origin[0] == 0:
        for e in row:
            e.x -= e.bounds[0] * len(row) / 2 - e.bounds[0] / 2

if __name__ == '__main__':
    app = Ursina()

    center = Entity(model='quad', scale=.1, color=color.red)
    p = Entity()
    for i in range(4*5):
        b = Button(parent=p, model='quad', scale=.5, scale_x=1, text=str(i), color=color.tint(color.random_color(),-.6))
        b.text_entity.world_scale = 1
    t = time.time()
    grid_layout(p.children, max_x=7, max_y=10, origin=(0, .5))
    center = Entity(parent=camera.ui, model=Circle(), scale=.005, color=color.lime)
    EditorCamera()
    print(time.time() - t)

    app.run()



from panda3d.core import Texture as PandaTexture
from panda3d.core import SamplerState
from panda3d.core import Vec2, Vec3, Vec4
from panda3d.core import Filename
from pathlib import Path
from direct.showbase import Loader
import sys
from ursina import color
# from PIL import Image


class Texture():

    default_filtering = None      # options: None / 'bilinear' / 'mipmap'

    def __init__(self, value):

        if isinstance(value, str):
            value = Path(value)

        if isinstance(value, Path):
            self.path = Path(value)
            self._texture = loader.loadTexture(Filename.fromOsSpecific(str(value)))
            self._cached_image = None   # for get_pixel() method

        elif isinstance(value, PandaTexture):
            self._texture = value

        else:
            from PIL import Image
            image = value
            self._texture = PandaTexture()
            self._texture.setup2dTexture(image.width, image.height, PandaTexture.TUnsignedByte, PandaTexture.FRgba)
            self._texture.setRamImageAs(image.transpose(Image.FLIP_TOP_BOTTOM).tobytes(), image.mode)
            self._cached_image = image.transpose(Image.FLIP_TOP_BOTTOM)
            self.path = None



        self.filtering = Texture.default_filtering      # None/'bilinear'/'mipmap' default: 'bilinear'


    @property
    def name(self):
        try:
            return self.path.name
        except:
            return f'PIL_texture_{self.size}'

    @property
    def size(self):
        return Vec2(self.width, self.height)

    @property
    def width(self):
        if self._cached_image:
            return self._cached_image.size[0]
        elif self._texture.getOrigFileXSize() > 0:
            return self._texture.getOrigFileXSize()
        return 0

    @property
    def height(self):
        if self._cached_image:
            return self._cached_image.size[1]
        elif self._texture.getOrigFileYSize() > 0:
            return self._texture.getOrigFileYSize()
        return 0

    @property
    def pixels(self):
        from numpy import asarray, flip
        from PIL import Image

        if self._cached_image:
            return asarray(self._cached_image)

        pixels = asarray(Image.open(self.path))
        pixels = flip(pixels, axis=0)
        return pixels


    @property
    def filtering(self):
        return self._filtering

    @filtering.setter
    def filtering(self, value):
        # print('setting filtering:', value)
        if value in (None, False, 'nearest', 'nearest neighbor', 'point'):
            self._texture.setMagfilter(SamplerState.FT_nearest)
            self._texture.setMinfilter(SamplerState.FT_nearest)
            self._filtering = False
        elif value in (True, 'linear', 'bilinear'):
            self._texture.setMagfilter(SamplerState.FT_linear)
            self._texture.setMinfilter(SamplerState.FT_linear)
            self._filtering = True
        elif value == 'mipmap':
            self._texture.setMinfilter(SamplerState.FT_linear_mipmap_linear)
            self._filtering = 'mipmap'



    def get_pixel(self, x, y):
        try:
            if not self._cached_image:
                from PIL import Image
                self._cached_image = Image.open(self.path)


            col = self._cached_image.getpixel((x, self.height-y-1))
            if self._cached_image.mode == 'LA':
                col = (col[0], col[0], col[0], col[1])

            if self._cached_image.mode == 'L':
                col = (col[0], col[0], col[0])

            return color.rgba(*col)
        except Exception as e:
            print(e)
            return None


    def get_pixels(self, start, end):
        start = (clamp(start[0], 0, self.width), clamp(start[1], 0, self.width))
        end = (clamp(end[0], 0, self.width), clamp(end[1], 0, self.width))
        pixels = list()

        for y in range(start[1], end[1]):
            for x in range(start[0], end[0]):
                pixels.append(self.get_pixel(x,y))

        return pixels

    def set_pixel(self, x, y, color):
        if not self._cached_image:
            from PIL import Image
            self._cached_image = Image.open(self.path)

        self._cached_image.putpixel((x, self.height-y-1), tuple([int(e*255) for e in color]))

    def apply(self):
        from PIL import Image
        if not self._cached_image:
            self._cached_image = Image.open(self.path)

        self._texture.setRamImageAs(self._cached_image.transpose(Image.FLIP_TOP_BOTTOM).tobytes(), self._cached_image.mode)
        # self._texture.setRamImageAs(self._cached_image.tobytes(), self._cached_image.mode)

    def save(self, path):
        if not self._cached_image:
            from PIL import Image
            self._cached_image = Image.open(self.path)

        self._cached_image.save(path)

if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    '''
        The Texture class rarely used manually but usually instantiated
        when assigning a texture to an Entity
        texture = Texture(path / PIL.Image / panda3d.core.Texture)

        A texture file can be a .png, .jpg or .psd.
        If it's a .psd it and no compressed version exists, it will compress it automatically.
    '''
    e = Entity(model='quad', texture='brick')
    e.texture.set_pixel(0, 2, color.blue)
    e.texture.apply()

    for y in range(e.texture.height):
        for x in range(e.texture.width):
            if e.texture.get_pixel(x,y) == color.blue:
                print('found blue pixel at:', x, y)


    app.run()




class Scene(Entity):
    def __init__(self, **kwargs):
        super().__init__()
        Entity(parent=self, model='cube', position=Vec3(0.0,0.0,4.0), color=color.hsv(0.0,1.0,1.0))
        Entity(parent=self, model='plane', position=Vec3(0.0,0.0,0.0), color=color.hsv(120.0,1.0,1.0))
        Entity(parent=self, model='cube', position=Vec3(-1.2448334693908691,0.0,1.6809353828430176), color=color.hsv(30.0,1.0,1.0))




class Scene(Entity):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        Entity(parent=self, position=Vec3(-1.3587,0.032,1.0418), model='cube', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(0.7849,0.034,-1.1488), scale=Vec3(1.9744,1,1), model='cube', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(-2.2072,1.5801,-1.6211), scale=Vec3(2.0645,3.3151,2.0645), model='diamond', origin=Vec3(0,-0.5,0), ignore=True)
        Entity(parent=self, position=Vec3(3.0613,0.1833,-2.9419), scale=Vec3(2.0645,3.3151,2.0645), model='diamond', origin=Vec3(0,-0.5,0), ignore=True)



from pathlib import Path
from copy import copy
from ursina import application
from ursina.texture import Texture


imported_textures = dict()
file_types = ('.tif', '.jpg', '.jpeg', '.png', '.gif')
textureless = False


def load_texture(name, path=None):
    if textureless:
        return None

    if name in imported_textures:
        return copy(imported_textures[name])

    folders = ( # folder search order
        application.compressed_textures_folder,
        application.asset_folder,
        application.internal_textures_folder,
        )
    if path:
        if isinstance(path, str):
            folders = (Path(path),)
        else:
            folders = (path,)

    if name.endswith('.mp4'):
        for folder in folders:
            for filename in folder.glob('**/' + name):
                # print('loaded movie texture:', filename)
                return loader.loadTexture(filename.resolve())


    for folder in folders:
        if '.' in name: # got name with file extention
            for filename in folder.glob('**/' + name):
                t = Texture(filename.resolve())
                imported_textures[name] = t
                return t

        for filename in folder.glob('**/' + name + '.*'): # no file extention given, so try all supported
            if filename.suffix in file_types:
                # print('found:', filename)
                t = Texture(filename.resolve())
                imported_textures[name] = t
                return t

    if application.development_mode:
        try:
            from psd_tools import PSDImage
        except (ModuleNotFoundError, ImportError) as e:
            print('info: psd-tools3 not installed')

        for folder in folders:
            for filename in folder.glob('**/' + name + '.psd'):
                print('found uncompressed psd, compressing it...')
                compress_textures(name)
                return load_texture(name)

    return None



def compress_textures(name=''):
    import os
    try:
        from PIL import Image
    except Exception as e:
        return e


    if not application.compressed_textures_folder.exists():
        application.compressed_textures_folder.mkdir()


    file_type = '.*'
    if '.' in name:
        file_type = ''

    # print('searching for texture:', name + file_type)
    for f in application.asset_folder.glob('**/' + name + file_type):

        if '\\textures_compressed\\' in str(f) or f.suffix not in ('.psd', '.png', '.jpg', '.jpeg', '.gif'):
            continue
        # print('  found:', f)
        image = None
        if f.suffix == '.psd':
            try:
                from psd_tools import PSDImage
            except (ModuleNotFoundError, ImportError) as e:
                print('info: psd-tools3 not installed')
                return None

            image = PSDImage.load(f)
            image = image.as_PIL()
        # elif f.suffix == '.png':
        #     image = Image.open(f)

        if not image:
            return False
        # print(max(image.size))
        # print('............', image.mode)
        if image and image.mode != 'RGBA' and max(image.size) > 512:
            image.save(
                application.compressed_textures_folder / (Path(f).stem + '.jpg'),
                'JPEG',
                quality=80,
                optimize=True,
                progressive=True
                )
            print('    compressing to jpg:', application.compressed_textures_folder / (f.stem + '.jpg'))
            continue
        else:
            image.save(application.compressed_textures_folder / (f.stem + '.png'), 'PNG')
            print('    compressing to png:', application.compressed_textures_folder / (f.stem + '.png'))



if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    Entity(model='quad', texture='white_cube')
    app.run()



from ursina import *
from collections import namedtuple
from copy import copy


class Node:
    __slots__ = ['index', 'indent_level', 'content', 'code', 'children', 'is_answer']

    def __str__(self):
        return 'Node:\n    ' + '\n    '.join([f'{e} = {getattr(self, e)}' for e in Node.__slots__])


bar_mission_solved = False
class Conversation(Entity):

    def __init__(self, **kwargs):
        super().__init__(parent=camera.ui, y=-.1)

        self.question = Button(
            parent=self,
            text_origin=(-.5,0),
            scale_x=1,
            scale_y=.1,
            model=Quad(radius=.5, aspect=1/.1),
            text='What do you want\nWhat do you want?'
            )
        self.question.text_entity.line_height = 1.25
        self.question.text_entity.position = (-.45, -.05)
        self.question.highlight_color = self.question.color
        self.more_indicator = Entity(parent=self.question, model=Circle(3), position=(.45,-.4,-.1), rotation_z=180, color=color.azure, world_scale=.5, z=-1, enabled=False)
        def toggle():
            self.more_indicator.visible = not self.more_indicator.visible
            invoke(self.more_indicator.toggle, delay=.5)

        self.more_indicator.toggle = toggle
        self.more_indicator.toggle()
        # self.question.text_entity.origin = (-.5,.5)
        self.spacing = 4 * .02
        self.wordwrap = 65
        self.button_model = Quad(radius=.5, aspect=1/.075)

        for key, value in kwargs.items():
            setattr(self, key ,value)

        self.answer_0 = Button(parent=self, text='answer_0', y=self.question.y-self.spacing-.025, scale=(1,.075), text_origin=(-.5,0), model=copy(self.button_model))
        self.answer_1 = Button(parent=self, text='answer_1', y=self.answer_0.y-self.spacing, scale=(1,.075), text_origin=(-.5,0), model=copy(self.button_model))
        self.answer_2 = Button(parent=self, text='answer_2', y=self.answer_1.y-self.spacing, scale=(1,.075), text_origin=(-.5,0), model=copy(self.button_model))

        self.buttons = (self.answer_0, self.answer_1, self.answer_2)
        for b in self.buttons:
            b.text_entity.line_height = 1.15
            b.text_entity.position = (-.45, 0)

        self.question_appear_sequence = None
        self.button_appear_sequence = None
        self.started = False



    def ask(self, node, question_part=0):
        # print(node)
        self.current_node = node
        self.question_part = question_part
        self.question.text = node.content[question_part]
        self.question.text_entity.wordwrap = self.wordwrap
        self.more_indicator.enabled = False
        self.question_appear_sequence = self.question.text_entity.appear(delay=.1)

        for b in self.buttons:
            b.enabled = False

        answers = list()
        for i, child in enumerate(node.children):
            # print(child)
            if child.code and child.code.startswith('if'):
                a = eval(child.code[3:])
                if a:
                    print('wooooooooooooooooooo', a)

            answers.append(child)

        # multi page question
        if len(node.content) > 1 and self.question_part < len(node.content)-1:
            if self.question_part < len(node.content): # question not finished, so don't show answer buttons
                # print('question not finished')
                self.question_appear_sequence.append(Func(setattr, self.more_indicator, 'enabled', True))
                return

        self.button_appear_sequence = Sequence()
        invoke(self.button_appear_sequence.start, delay=self.question_appear_sequence.duration)

        print('aaaaaa', [n.content for n in node.children])
        if not node.children:
            self.buttons[0].text = '*leave*'
            self.buttons[0].on_click = Func(setattr, self, 'enabled', False)
            self.button_appear_sequence.append(Func(setattr, self.buttons[0], 'enabled', True))


        for i, child in enumerate(node.children):
            self.button_appear_sequence.append(Wait(i*.15))
            self.button_appear_sequence.append(Func(setattr, self.buttons[i], 'enabled', True))
            self.buttons[i].text = child.content[0]
            self.buttons[i].text_entity.wordwrap = self.wordwrap

            def on_click(node=child):
                # print(node)
                if not node.children:
                    print('end')
                    self.enabled = False
                    return


                invoke(self.ask, node.children[0], 0, delay=.1)
                if len(node.children) > 1:
                    print('error at node:', n, '. node has multiple children, but should only have one (a question)')

            self.buttons[i].on_click = on_click



    def input(self, key):
        if key == 'left mouse down' or key == 'space' and not mouse.hovered_entity in self.buttons:
            self.next()


    def next(self):
        if not self.started:
            return

        if not self.question_appear_sequence.finished and self.question_appear_sequence.t > .1:
            self.question_appear_sequence.finish()
            if self.button_appear_sequence:
                self.button_appear_sequence.start()
            return

        if self.question_part < len(self.current_node.content)-1:
            self.ask(self.current_node, self.question_part+1)


    def start_conversation(self, conversation):
        self.conversation_nodes = self.parse_conversation(conversation)
        self.ask(self.conversation_nodes[0])
        self.started = True



    def parse_conversation(self, convo):
        convo = convo.strip()
        nodes = list()
        prev_node = None
        node_index = 0

        for i, l in enumerate(convo.split('\n')):
            # if i <= 1 and l.startswith('('):
            #     npc_name = l[1:-1]
            #     print('name: ' + npc_name)
            #     continue
            if not l:
                continue

            indent_level = len(l) - len(l.lstrip())
            indent_level //= 4
            content, code = l.strip(), None
            if '(' in l:
                content, code = l.split('(')
                content = content.strip()
                code = code[:-1]
                is_answer = content.startswith('* ')
                # print('code:', codeK)

            if prev_node and prev_node.indent_level == indent_level:
                prev_node.content.append(content)
                prev_node = n
                continue

            n = Node()
            n.index = node_index
            n.indent_level = indent_level
            n.is_answer = content.startswith('* ')
            if n.is_answer:
                content = content[2:]
            n.content = [content, ]
            n.children = list()
            n.code = code
            nodes.append(n)
            prev_node = n
            node_index += 1
            # if i == 0:
            #     continue

            # look backwards through nodes to find current node's parent
            for j in range(node_index-1, -1, -1):
                if nodes[j].indent_level == n.indent_level-1:
                    nodes[j].children.append(n)
                    break


        # for n in nodes:
        #     print(n.content)
        #     for c in n.children:
        #         print(' |__', c.content)
        #     if not n.children:
        #         print(' |__leave')

        return nodes



if __name__ == '__main__':
    app = Ursina()
    # import main
    # Button.color = color.hsv(0,0,0,.8)
    conversation = Conversation()
    # conversation.question.model = 'quad'
    # for b in conversation.buttons:
    #     b.model = 'quad'
    bar_mission_solved = False
    convo = '''
I'm looking for my sister. Can you help me find her, please? I haven't seen her in days! Who know what could've happened!?
I'm worried. Will you help me?
    * Yes, of course. This can be a dangerous city.
        Oh no! Do you think something happened to her?
        What should I do?!
            * She's probably fine. She can handle herself.
                You're right. I'm still worried though.
                    * Don't worry, I'll look for her.
            * Maybe. (chaos += 1)
                Help me look for her, please! *runs off*
    * I'm sorry, but I don't have time right now.
        A true friend wouldn't say that. (evil += 1)
    * I know where she is! (if bar_mission_solved)
        Really? Where?
            * I saw her on a ship by the docks, it looked like they were ready to set off.
                Thank you! *runs off*
'''
    conversation.start_conversation(convo)
    # conversation.parse_conversation(convo)

    # def input(key):
    #     if key == 'space':
    #         conversation.start_conversation()
    # window.color = color._16
    window.size = window.fullscreen_size * .5
    Sprite('shore', z=1)
    app.run()



from ursina import *; unlit_shader = Shader(language=Shader.GLSL, vertex = '''#version 140


uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 texcoords;
uniform vec2 texture_scale;
uniform vec2 texture_offset;

void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    texcoords = (p3d_MultiTexCoord0 * texture_scale) + texture_offset;
}
''',

fragment='''
#version 140

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 texcoords;
out vec4 fragColor;

void main() {
    vec4 color = texture(p3d_Texture0, texcoords) * p3d_ColorScale;
    fragColor = color.rgba;
}

''',
default_input={
    'texture_scale' : Vec2(1,1),
    'texture_offset' : Vec2(.5,0),
}
)



if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina()
    # window.color=color.black
    # from ursina.lights import DirectionalLight
    # DirectionalLight()

    shader = unlit_shader

    # a = AzureCube(shader=shader)
    # b = YellowSphere(shader=shader, rotation_y=180, x=3, texture='shore')
    # from panda3d.core import Material
    # myMaterial = Material()
    # myMaterial.setShininess(5.0) #Make this material shiny
    # myMaterial.setAmbient((0, 0, 1, 1)) #Make this material blue
    # b.set_material(myMaterial)
    # AzureSphere(shader=a.shader, y=2)
    ground = GrayPlane(scale=10, y=-2, texture='shore', shader=shader, texture_scale=(10,10))
    ground.set_shader_input('texture_scale', Vec2(2, 1))
    #Sky(color=color.light_gray)
    EditorCamera()

    app.run()



from ursina import *


class DropdownMenuButton(Button):
    def __init__(self, text='', **kwargs):
        super().__init__(
            scale=(.25,.025),
            origin=(-.5,.5),
            pressed_scale=1,
            text=text,
            **kwargs
            )

        if self.text_entity:
            self.text_entity.x = .05
            self.text_entity.origin = (-.5, 0)
            self.text_entity.scale *= .8


class DropdownMenu(DropdownMenuButton):

    def __init__(self, text='', buttons=list(), **kwargs):
        super().__init__(text=text)
        self.position = window.top_left
        self.buttons = buttons
        for i, b in enumerate(self.buttons):
            b.world_parent = self
            b.original_scale = b.scale
            b.x = 0
            b.y = -i-1 *.98
            b.enabled = False

            if isinstance(b, DropdownMenu):
                for e in b.buttons:
                    e.x = 1
                    e.y += 1

        self.arrow_symbol = Text(world_parent=self, text='>', origin=(.5,.5), position=(.95, 0), color=color.gray)
        for key, value in kwargs.items():
            setattr(self, key, value)


    def open(self):
        for i, b in enumerate(self.buttons):
            invoke(setattr, self.buttons[i], 'enabled', True, delay=(i*.02))

    def close(self):
        for i, b in enumerate(reversed(self.buttons)):
            b.enabled = False


    def on_mouse_enter(self):
        super().on_mouse_enter()
        self.open()

    def input(self, key):
        if key == 'left mouse down' and mouse.hovered_entity and mouse.hovered_entity.has_ancestor(self):
            self.close()

    def update(self):
        if self.hovered or mouse.hovered_entity and mouse.hovered_entity.has_ancestor(self):
            return

        self.close()


if __name__ == '__main__':
    from ursina.prefabs.dropdown_menu import DropdownMenu, DropdownMenuButton

    app = Ursina()
    # DropdownMenu(text='File')
    DropdownMenu('File', buttons=(
        DropdownMenuButton('New'),
        DropdownMenuButton('Open'),
        DropdownMenu('Reopen Project', buttons=(
            DropdownMenuButton('Project 1'),
            DropdownMenuButton('Project 2'),
            )),
        DropdownMenuButton('Save'),
        DropdownMenu('Options', buttons=(
            DropdownMenuButton('Option a'),
            DropdownMenuButton('Option b'),
            )),
        DropdownMenuButton('Exit'),
        ))

    app.run()



from ursina import *; lit_with_shadows_shader = Shader(language=Shader.GLSL, vertex = '''#version 150
uniform struct {
  vec4 position;
  vec3 color;
  vec3 attenuation;
  vec3 spotDirection;
  float spotCosCutoff;
  float spotExponent;
  sampler2DShadow shadowMap;
  mat4 shadowViewMatrix;
} p3d_LightSource[1];

const float M_PI = 3.141592653589793;


uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelViewMatrix;
uniform mat3 p3d_NormalMatrix;

in vec4 vertex;
in vec3 normal;

in vec2 p3d_MultiTexCoord0;
uniform vec2 texture_scale;
uniform vec2 texture_offset;
out vec2 texcoords;


out vec3 vpos;
out vec3 norm;
out vec4 shad[1];

void main() {
  gl_Position = p3d_ModelViewProjectionMatrix * vertex;
  vpos = vec3(p3d_ModelViewMatrix * vertex);
  norm = normalize(p3d_NormalMatrix * normal);
  shad[0] = p3d_LightSource[0].shadowViewMatrix * vec4(vpos, 1);
  texcoords = (p3d_MultiTexCoord0 * texture_scale) + texture_offset;
}

''',
fragment='''
#version 150
uniform struct {
  vec4 position;
  vec3 color;
  vec3 attenuation;
  vec3 spotDirection;
  float spotCosCutoff;
  float spotExponent;
  sampler2DShadow shadowMap;
  mat4 shadowViewMatrix;
} p3d_LightSource[1];

const float M_PI = 3.141592653589793;


uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 texcoords;

uniform struct {
  vec4 ambient;
} p3d_LightModel;

uniform struct {
  vec4 ambient;
  vec4 diffuse;
  vec3 specular;
  float roughness;
} p3d_Material;

in vec3 vpos;
in vec3 norm;
in vec4 shad[1];

out vec4 p3d_FragColor;
uniform vec4 shadow_color;

void main() {
  p3d_FragColor = texture(p3d_Texture0, texcoords) * p3d_ColorScale;

  // float alpha = p3d_Material.roughness * p3d_Material.roughness;
  vec3 N = norm;

  for (int i = 0; i < p3d_LightSource.length(); ++i) {
    vec3 diff = p3d_LightSource[i].position.xyz - vpos * p3d_LightSource[i].position.w;
    vec3 L = normalize(diff);
    vec3 V = normalize(-vpos);
    vec3 H = normalize(L + V);

    float NdotL = clamp(dot(N, L), 0.001, 1.0);
    // float NdotV = clamp(abs(dot(N, V)), 0.001, 1.0);
    // float NdotH = clamp(dot(N, H), 0.0, 1.0);
    // float VdotH = clamp(dot(V, H), 0.0, 1.0);

    // Specular term
    // float reflectance = max(max(p3d_Material.specular.r, p3d_Material.specular.g), p3d_Material.specular.b);
    // float reflectance90 = clamp(reflectance * 25.0, 0.0, 1.0);
    // vec3 F = p3d_Material.specular + (vec3(reflectance90) - reflectance) * pow(clamp(1.0 - VdotH, 0.0, 1.0), 5.0);

    // Geometric occlusion term
    // float alpha2 = alpha * alpha;
    // float attenuationL = 2.0 * NdotL / (NdotL + sqrt(alpha2 + (1.0 - alpha2) * (NdotL * NdotL)));
    // float attenuationV = 2.0 * NdotV / (NdotV + sqrt(alpha2 + (1.0 - alpha2) * (NdotV * NdotV)));
    // float G = attenuationL * attenuationV;

    // Microfacet distribution term
    // float f = (NdotH * alpha2 - NdotH) * NdotH + 1.0;
    // float D = alpha2 / (M_PI * f * f);

    // Lambert, energy conserving
    // vec3 diffuseContrib = (1.0 - F) * p3d_Material.diffuse.rgb / M_PI;
    vec3 diffuseContrib = p3d_Material.diffuse.rgb / M_PI;

    // Cook-Torrance
    //vec3 specContrib = F * G * D / (4.0 * NdotL * NdotV);
    // vec3 specContrib = vec3(0.);

    // Obtain final intensity as reflectance (BRDF) scaled by the energy of the light (cosine law)
    // vec3 color = NdotL * p3d_LightSource[i].color * (diffuseContrib + specContrib);
    vec3 color =  NdotL * p3d_LightSource[i].color * diffuseContrib;
    const float bias = 0.001;

    vec4 shadowcoord = shad[i];
    shadowcoord.z += bias;

    vec3 converted_shadow_color = (vec3(1.,1.,1.) - shadow_color.rgb) * shadow_color.a;

    p3d_FragColor.rgb += textureProj(p3d_LightSource[i].shadowMap, shadowcoord) * converted_shadow_color;
    p3d_FragColor.rgb += color - converted_shadow_color;
  }
}

''',
default_input = {
    'texture_scale': Vec2(1,1),
    'texture_offset': Vec2(0,0),
    'shadow_color' : Vec4(0, .5, 1, .25),
    }
)



if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    # window.size *= .5
    app = Ursina()
    shader = lit_with_shadows_shader

    a = Entity(model='cube', shader=shader, color=color.light_gray)
    Entity(model='cube', color=color.white, scale=(10,1,10), y=-2, shader=shader, texture='white_cube')
    Entity(model='cube', color=color.gray, scale=(5,1,10), x=-3.5, y=-1, shader=shader)


    # Enable shadows; we need to set a frustum for that.
    from ursina.lights import DirectionalLight
    # sun = DirectionalLight(y=10, rotation=(90+30,90,0))
    sun = DirectionalLight()
    sun.look_at(Vec3(-1,-1,-1))
    # sun._light.show_frustum()

    Sky(color=color.light_gray)
    EditorCamera()

    def update():
        a.x += (held_keys['d'] - held_keys['a']) * time.dt * 5
        a.y += (held_keys['e'] - held_keys['q']) * time.dt * 5
        a.z += (held_keys['w'] - held_keys['s']) * time.dt * 5


    app.run()



from ursina import *; fading_shadows_shader = Shader(language=Shader.GLSL, vertex = '''#version 150

uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelViewMatrix;
uniform mat3 p3d_NormalMatrix;

in vec4 vertex;
in vec3 normal;

in vec2 p3d_MultiTexCoord0;
out vec2 texcoords;

out vec3 vpos;
out vec3 norm;


// uniform float4x4 trans_model_to_clip_of_light,
// #              uniform float4x4 mat_modelproj,
// #              uniform float4 mspos_light,
// #              uniform float4 k_ambient,
// #              uniform float4 k_scale,
// #              uniform float4 k_push,
// #
// #              out float4 l_position : POSITION,
// #              out float2 l_texcoord0 : TEXCOORD0,
// #              out float4 l_shadowcoord : TEXCOORD1,
// # 	     out float  l_smooth : TEXCOORD2,
// # 	     out float4 l_lightclip : TEXCOORD3

void main() {
  gl_Position = p3d_ModelViewProjectionMatrix * vertex;
  vpos = vec3(p3d_ModelViewMatrix * vertex);
  norm = normalize(p3d_NormalMatrix * normal);
  texcoords = p3d_MultiTexCoord0;
}
''',
fragment='''
in vec3 vpos;
in vec3 norm;

out vec4 p3d_FragColor;

void main() {
  p3d_FragColor = texture(p3d_Texture0, texcoords) * p3d_ColorScale;


  }
  p3d_FragColor.a = 1;
}


''',
)
if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina()
    shader = fading_shadows_shader

    a = AzureCube(shader=shader, texture='shore')
    b = WhiteSphere(shader=shader, rotation_y=180, x=3, texture='brick')
    b.texture.filtering = None
    GrayPlane(scale=10, y=-2, texture='shore', shader=shader)

    Sky(color=color.light_gray)
    EditorCamera()

    def update():
        a.x += held_keys['d'] * .1
        a.x -= held_keys['a'] * .1
        a.y += held_keys['e'] * .1
        a.y -= held_keys['q'] * .1


    app.run()


# //Cg
#
# void vshader(float4 vtx_position : POSITION,
#              float2 vtx_texcoord0: TEXCOORD0,
#              float3 vtx_normal: NORMAL,
#
#              uniform float4x4 trans_model_to_clip_of_light,
#              uniform float4x4 mat_modelproj,
#              uniform float4 mspos_light,
#              uniform float4 k_ambient,
#              uniform float4 k_scale,
#              uniform float4 k_push,
#
#              out float4 l_position : POSITION,
#              out float2 l_texcoord0 : TEXCOORD0,
#              out float4 l_shadowcoord : TEXCOORD1,
# 	     out float  l_smooth : TEXCOORD2,
# 	     out float4 l_lightclip : TEXCOORD3
#              )
#
# {
# float4 position = vtx_position * k_scale;
#
# // vertex position
# l_position = mul(mat_modelproj, position);
#
# // Pass through texture coordinate for main texture.
# l_texcoord0 = vtx_texcoord0;
#
# // Calculate the surface lighting factor.
# l_smooth = saturate(dot(vtx_normal, normalize(mspos_light - position)));
#
# // Calculate light-space clip position.
# float4 pushed = position + float4(vtx_normal * k_push, 0);
# l_lightclip = mul(trans_model_to_clip_of_light, pushed);
#
# // Calculate shadow-map texture coordinates.
# l_shadowcoord = l_lightclip * float4(0.5,0.5,0.5,1.0) + l_lightclip.w * float4(0.5,0.5,0.5,0.0);
# }
#
#
# void fshader(in float2 l_texcoord0 : TEXCOORD0,
#              in float4 l_shadowcoord : TEXCOORD1,
#              in float  l_smooth : TEXCOORD2,
#              in float4 l_lightclip : TEXCOORD3,
#              uniform sampler2D tex_0 : TEXUNIT0,
#              uniform sampler2D k_Ldepthmap : TEXUNIT1,
#              uniform float4 k_ambient,
# 	     uniform float4 k_texDisable,
#              out float4 o_color:COLOR)
# {
#   float3 circleoffs = float3(l_lightclip.xy / l_lightclip.w, 0);
#   float falloff = saturate(1.0 - dot(circleoffs, circleoffs));
#   float4 baseColor = saturate(tex2D(tex_0, l_texcoord0) + k_texDisable);
#   float shade = tex2Dproj(k_Ldepthmap,l_shadowcoord);
#
#   /***** this is the only line that has changed *****/
#   o_color = baseColor * ( falloff * shade * l_smooth + (1.0-falloff)*l_smooth + k_ambient.x );
# }



from ursina import *

class DebugMenu(Draggable):
    def __init__(self, target, **kwargs):
        super().__init__()
        self.target = target
        self.scale = (.2, .025)
        self.draw_functions()
        self.text = '<orange>' + target.__class__.__name__

        for key, value in kwargs.items():
            setattr(self, key, value)


    def draw_functions(self):
        for c in self.children:
            destroy(c)
        for i, f in enumerate([func for func in self.target.__class__.__dict__
        if callable(getattr(self.target.__class__, func))
        and not func.startswith("__")]):
            # print('functions:', f)
            b = Button(
                parent = self,
                text = f + '()',
                y = -i - 1,
                on_click = getattr(self.target, f)
                )
            b.text_entity.world_scale = 1


if __name__ == '__main__':
    app = Ursina()
    DebugMenu(Audio('night_sky'))
    app.run()



import colorsys
import random
import sys
from math import floor, inf
from panda3d.core import Vec4


class Color(Vec4):

    def __init__(self,*p):
        super().__init__(*p)

    def __str__(self):
        return f'Color({self[0]}, {self[1]}, {self[2]}, {self[3]})'

    @property
    def name(self):
        for key, value in colors.items():
            if value == self:
                return key

        return None

    @property
    def r(self):
        return self[0]
    @property
    def g(self):
        return self[1]
    @property
    def b(self):
        return self[2]
    @property
    def a(self):
        return self[3]

    @property
    def hsv(self):
        result = to_hsv((self[0], self[1], self[2], self[3]))
        result[0] =  floor(result[0] * 360)
        return result

    @property
    def h(self):
        return self.hsv[0]
    @property
    def s(self):
        return self.hsv[1]
    @property
    def v(self):
        return self.hsv[2]

    @property
    def brightness(self):
        return brightness(self)


    def invert(self):
        return inverse(self)

    def tint(self, amount):
        return tint(self, amount)



def hsv(h, s, v, a=1):
    return Color(colorsys.hsv_to_rgb((h / 360) - floor(h / 360), s, v) + (a,))

color = hsv

def rgba(r, g, b, a=255):
    color = Color(r, g, b, a)
    if color[0] > 1 or color[1] > 1 or color[2] > 1 or color[3] > 1:
        color = Color(tuple(c/255 for c in color))
    # color[3] = min(1, color[3])
    return color

def rgb(r, g, b, a=255):
    return rgba(r, g, b, a)

def to_hsv(color):
    return Color(colorsys.rgb_to_hsv(color[0], color[1], color[2]) + (color[3],))

def hex(value):
    if value.startswith('#'):
        value = value[1:]
    return rgb(*tuple(int(value[i:i+2], 16) for i in (0, 2, 4)))

def brightness(color):
    if color[0] > 1 or color[1] > 1 or color[2] > 1:
        color = Color(tuple(c/255 for c in color))
    return to_hsv(color)[2]

def inverse(color):
    color = Color(tuple(1 - c for c in color))
    color[3] = 1
    return color

def random_color():
    return Color(random.random(), random.random(), random.random(), 1)

def tint(color, amount=.2):
    return Color(
        max(min(color[0] + amount, 1), 0),
        max(min(color[1] + amount, 1), 0),
        max(min(color[2] + amount, 1), 0),
        color[3]
        )


white =         color(0, 0, 1)
smoke =         color(0, 0, 0.96)
light_gray =    color(0, 0, 0.75)
gray =          color(0, 0, 0.5)
dark_gray =     color(0, 0, 0.25)
black =         color(0, 0, 0)
red =           color(0, 1, 1)
orange =        color(30, 1, 1)
yellow =        color(60, 1, 1)
lime =          color(90, 1, 1)
green =         color(120, 1, 1)
turquoise =     color(150, 1, 1)
cyan =          color(180, 1, 1)
azure =         color(210, 1, 1)
blue =          color(240, 1, 1)
violet =        color(270, 1, 1)
magenta =       color(300, 1, 1)
pink =          color(330, 1, 1)

brown =         rgb(165, 42, 42)
olive =         rgb(128, 128, 0)
peach =         rgb(255, 218, 185)
gold =          rgb(255, 215, 0)
salmon =        rgb(250, 128, 114)

clear =         Color(0, 0, 0, 0)
white10 =       Color(1,1,1, 0.10)
white33 =       Color(1,1,1, 0.33)
white50 =       Color(1,1,1, 0.50)
white66 =       Color(1,1,1, 0.66)
black10 =       Color(0,0,0, 0.10)
black33 =       Color(0,0,0, 0.33)
black50 =       Color(0,0,0, 0.50)
black66 =       Color(0,0,0, 0.66)
black90 =       Color(0,0,0, 0.90)

text = smoke
light_text = smoke
dark_text = color(0, 0, .1)
text_color = light_text

# generate attributes for grayscale values. _0=black, _128=gray, 255=white
for i in range(256):
    setattr(sys.modules[__name__], '_' + str(i), color(0,0,i/255))

color_names = ('white', 'smoke', 'light_gray', 'gray', 'dark_gray', 'black',
    'red', 'orange', 'yellow', 'lime', 'green', 'turquoise', 'cyan', 'azure',
    'blue', 'violet', 'magenta', 'pink', 'brown', 'olive', 'peach', 'gold', 'salmon')
colors = dict()
for cn in color_names:
    colors[cn] = getattr(sys.modules[__name__], cn)


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    print(color.brightness(color.blue))
    print(_3)

    p = Entity(x=-2)
    for key in color.colors:
        print(key)
        b = Button(parent=p, model=Quad(subdivisions=2), color=color.colors[key], text=key)
        b.text_entity.scale *= .5

    grid_layout(p.children, max_x=8)

    for name in ('r', 'g', 'b', 'h', 's', 'v', 'brightness'):
        print(name + ':', getattr(color.random_color(), name))

    e = Entity(model='cube', color=color.lime)
    print(e.color.name)

    # e.color = hex('ced9a9')

    app.run()



from ursina import *
from math import floor

app = Ursina()
t = time.time()

application.asset_folder = application.asset_folder.parent.parent
terrain = Entity(model=Terrain('grass_fields_heightmap', skip=8), texture='grass', texture_scale=(3,3), scale=256)


# grid = [[[None for z in range(8)] for y in range(1)] for x in range(8)] # make 2d array of entities
grid = [[None for z in range(8)] for x in range(8)] # make 2d array of entities

x_slices = 8
# y = 1
z_slices = 8

for z in range(z_slices):
    for x in range(x_slices):
        part = Entity(
            parent=terrain,
            x=(x/x_slices) - .5 + (1/x_slices/2),
            z=(z/z_slices) - .5 + (1/z_slices/2),
            color=color.random_color(),
            model=Mesh(),
            always_on_top=True
            )
        grid[x][z] = part

terrain.model.generated_vertices = [v+Vec3(.5,0.5) for v in terrain.model.generated_vertices]
for i in range(0, len(terrain.model.generated_vertices), 3):
    v = terrain.model.generated_vertices[i]
    x = floor(v.x * x_slices)
    z = floor(v.z * z_slices)
    x = min(x, x_slices-1)
    z = min(z, z_slices-1)

    offset = Vec3(- (x/x_slices) - (1/x_slices/2), -.5, -(z/z_slices) - (1/x_slices/2))
    grid[x][z].model.vertices.extend([
        terrain.model.generated_vertices[i]   + offset,
        terrain.model.generated_vertices[i+1] + offset,
        terrain.model.generated_vertices[i+2] + offset,
        ])


for z in range(z_slices):
    for x in range(x_slices):
        grid[x][z].model.generate()
        # Entity(parent=grid[x][z], model='cube', scale=.01, color=color.red, always_on_top=True)
        # grid[x][z].enabled = False

        grid[x][z].collider = 'mesh'
        # grid[x][z].model = None


from ursina.prefabs.first_person_controller import FirstPersonController
player = FirstPersonController(position=(0,200,0))
player.add_script(NoclipMode())
# player = EditorCamera(rotation_x=90, y=128)


def update():
    for part in terrain.children:
        part.enabled = distance_xz(part.world_position, player.world_position) < 256/8
        # print(distance_xz(part.world_position, camera.world_position), 256/4)






app.run()



from ursina import *


class Cone(Mesh):
    def __init__(self, resolution=4, radius=.5, height=1, direction=(0,1,0), add_bottom=True, mode='triangle', **kwargs):
        origin = Entity()
        point = Entity(parent=origin)

        axis_index = direction.index(1)
        right = axis_index+1
        if right >= 3:
            right = 0
        setattr(point, ('x', 'y', 'z')[right], radius)

        rot_axis = 'rotation_' + ('x', 'y', 'z')[axis_index]

        verts = list()
        for i in range(resolution):
            verts.append(point.world_position) # outer point
            if add_bottom:
                verts.append((0,0,0)) #center point

            setattr(origin, rot_axis, getattr(origin, rot_axis) - 360/resolution)
            verts.append(point.world_position) # next outer point

            if add_bottom:
                setattr(origin, rot_axis, getattr(origin, rot_axis) - 360/resolution)

            verts.append((height*direction[0], height*direction[1], height*direction[2])) # spike point

            if add_bottom:
                verts.append(point.world_position)
                setattr(origin, rot_axis, getattr(origin, rot_axis) - 360/resolution)
                verts.append(point.world_position)

        destroy(origin)
        destroy(point)

        super().__init__(vertices=verts, mode=mode, **kwargs)


if __name__ == '__main__':
    app = Ursina()
    # Entity(model=Cone(8), color=color.color(60,1,1,.3))
    Entity(model=Cone(8, direction=(0,1,0)), color=color.color(60,1,1,.3))
    origin = Entity(model='quad', color=color.orange, scale=(.05, .05))
    ed = EditorCamera(rotation_speed = 200, panning_speed=200)
    app.run()



from modulefinder import ModuleFinder
import os
import sys
import shutil
from shutil import copy, copyfile
from distutils.dir_util import copy_tree
from pathlib import Path
import time
from textwrap import dedent


def copytree(src, dst, symlinks=False, ignore=None):
    src = str(src)
    dst = str(dst)

    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            try:
                shutil.copytree(s, d, symlinks, ignore)
            except Exception as e:
                print(e)
        else:
            if s.endswith('.psd'):
                continue
            shutil.copy2(s, d)



project_folder = Path.cwd()
project_name = project_folder.stem
build_folder = Path(project_folder / 'build')
ignore = []
include_modules = []


for i, arg in enumerate(sys.argv):
    if arg == '--help':
        print(dedent('''
            package ursina application for windows10.
            provided with project folder path, creates a build folder where
            it copies python and project's dependent packages. requires a main.py file.
            copies game scripts and assets into 'build/scr' folder.
            creates a .bat file to start the game.
            include extra modules like this: --include_modules module_one,module_two,module_tree
            --ignore            # add assets to ignore
            --name              # change project name
            --include_modules   # inlude extra modules
            --overwrite         # don't ask to overwrite existing build, just overwrite
            '''
            )
        )
        sys.exit()

    elif arg == '--ignore':
        for j in range(i, len(sys.argv)):
            if sys.argv[j].startswith('-'):
                break
            ignore.append(sys.argv[j])
            print('ignoring', sys.argv[j])

    elif arg == '--name':
        project_name = sys.argv[i+1]

    elif arg == '--include_modules':
        include_modules = sys.argv[i+1].split(',')


if build_folder.exists():
    if not '--overwrite' in sys.argv:
        msg = f'Build folder {build_folder} already exists. \nProceed to delete and overwrite?'
        overwrite = input("%s (y/N) " % msg).lower() == 'y'
        if not overwrite:
            print('stopped building')
            exit()

    print('deleting existing build folder')
    shutil.rmtree(str(build_folder))

print('building project:', project_folder)
start_time = time.time()

python_dest = Path(build_folder / 'python')
python_dlls_dest = Path(build_folder / 'python/DLLs')
python_lib_dest = Path(build_folder / 'python/Lib')
src_dest = Path(build_folder / 'src')

build_folder.mkdir()
python_dest.mkdir()
python_dlls_dest.mkdir()
python_lib_dest.mkdir()
src_dest.mkdir()

# def copy_python():
# copy files in python installation folder, but not the folders
print('copying python')
python_folder = Path(sys.executable).parent
[copy(str(f), str(python_dest / f.name)) for f in python_folder.iterdir() if f.is_file()]


# def copy_python_lib():
print('copying python Lib files')
[copy(str(f), str(python_lib_dest / f.name)) for f in Path(python_folder / 'Lib').iterdir() if f.is_file()]


# def copy_always_included():
print('copying always included modules and extra modules')
include_modules = [f'Lib/site-packages/{e}' for e in include_modules]

always_include = [
    'Lib/collections',
    'Lib/ctypes',
    'Lib/encodings',
    'Lib/importlib',
    'Lib/urllib',
    'Lib/logging',
    'Lib/xml',
    'Lib/site-packages/panda3d/',
    'Lib/site-packages/screeninfo/',
    'Lib/site-packages/direct/',
    'DLLs/libffi-7.dll',
    'DLLs/_ctypes.pyd',
    ]

for path in always_include + include_modules:
    source = python_folder / path
    dest = python_dest / path
    print('copying always include:', source, '-->', dest)

    if source.is_file():
        dest.parent.mkdir(parents=True, exist_ok=True)
        copy(str(source), str(dest))
    elif source.is_dir():
        dest.mkdir(parents=True, exist_ok=True)
        copytree(source, dest)



# def copy_ursina():
print('copying ursina')
import importlib
spec = importlib.util.find_spec('ursina')
ursina_path = Path(spec.origin).parent
dest = build_folder / 'python/Lib/site-packages/ursina'
dest.mkdir(parents=True, exist_ok=True)
copytree(ursina_path, dest, ignore=shutil.ignore_patterns('samples', 'unused'))


# print('copying found modules')
# finder = ModuleFinder()
# finder.run_script(str(project_folder) + '/main.py')

# for name, mod in finder.modules.items():
#     filename = mod.__file__
#
#     if filename is None:
#         continue
#     if '__' in name:
#         print('ignore:', filename)
#         continue
#
#     if 'Python' in filename and 'DLLs' in filename:
#         print('copying:', filename)
#         copy(filename, str(build_folder / 'python/DLLs'))
#
#     elif 'lib\\site-packages\\' in filename:
#         print('copying module:', filename)
#         forward_slashed = filename.split('lib\\site-packages\\')[1].replace('\\', '/')
#         dir = build_folder / 'python/lib/site-packages' / forward_slashed
#         dir.parent.mkdir(parents=True, exist_ok=True)
#         copy(filename, dir)


print('copying assets')
for f in project_folder.iterdir():
    name = f.name
    dest = Path(src_dest / f.name)
    if name in ['.git', '__pycache__', 'build', '.gitignore'] + ignore:
        print('ignore:', f)
        continue
    elif f.is_dir():
        print('copying assetfolder:', f, 'to', dest)
        dest.mkdir(parents=True, exist_ok=True)
        copytree(project_folder / f, dest, ignore=shutil.ignore_patterns('*psd'))
    elif f.is_file():
        print('copying asset:', f, 'to', src_dest / f.name)
        copy(str(f), str(dest))


print('creating .bat file')
with Path(build_folder / f'{project_name}.bat').open('w') as f:
    f.write(dedent(f'''
        chcp 65001
        set PYTHONIOENCODING=utf-8

        call "%CD%\python\python.exe" "%CD%\src\main.py" > "log.txt" 2>&1'''
    ))
#
# # # make exe
# # import subprocess
# # subprocess.call([
# #     f'{ursina_path}\\scripts\\_bat_to_exe.bat',
# #     f'{build_folder}\\f'{project_name}.bat',
# #     f'\\build\\{project_folder.stem}.exe'
# #     ])
#

print('build complete! time elapsed:', time.time() - start_time)



from ursina import *


class HealthBar(Button):

    def __init__(self, max_value=100, show_text=True, show_lines=False, **kwargs):
        super().__init__(
            position = (-.45 * window.aspect_ratio, .45),
            origin = (-.5, .5),
            scale = (Text.size*20, Text.size),
            color = color.black66,
            highlight_color = color.black66,
            text = 'hp / max hp',
            ignore = True,
            )

        self.bar = Entity(parent=self, model='quad', origin=self.origin, z=-.01, color=color.red.tint(-.2), ignore=True)
        self.animation_duration = .1
        self.lines = Entity(parent=self.bar, origin=self.origin, y=-1, color=color.black33, ignore=True)
        self.text_entity.scale *= .7
        self.roundness = .25

        self.max_value = max_value
        self.clamp = True
        self.show_lines = show_lines
        self.show_text = show_text
        self.scale_x = self.scale_x # update rounded corners

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.scale_y = self.scale_y # update background's rounded corners
        self.value = self.max_value


    @property
    def value(self):
        return self._value


    @value.setter
    def value(self, n):
        if self.clamp:
            n = clamp(n, 0, self.max_value)

        self._value = n

        self.lines.enabled = False
        self.bar.animate_scale_x(n/self.max_value, duration=self.animation_duration, curve=curve.in_out_bounce)
        invoke(setattr, self.lines, 'enabled', True, delay=.1)

        self.text_entity.text = f'{n} / {self.max_value}'

        if self.lines.enabled:
            self.lines.model = Grid(n, 1)

        if n / self.max_value >= self.scale_y / self.scale_x:
            aspect_ratio = n/self.max_value*self.scale_x / self.scale_y
            self.bar.model = Quad(radius=self.roundness, aspect=aspect_ratio)
        else:
            self.bar.model = 'quad'
        self.bar.origin = self.bar.origin



    def __setattr__(self, name, value):
        if name == 'show_text':
            self.text_entity.enabled = value

        if name == 'show_lines':
            self.lines.enabled = value
            return

        if name == 'bar_color':
            self.bar.color = value
            return


        super().__setattr__(name, value)

        if 'scale' in name and hasattr(self, 'roundness'):  # update rounded corners of background when scaling
            orginal_text_position = self.text_entity.position
            self.model = Quad(radius=self.roundness, aspect=self.world_scale_x / self.world_scale_y)
            self.origin = self.origin
            self.text_entity.position = orginal_text_position

        if 'scale' in name and hasattr(self, 'text_entity'):  # make sure the text doesn't scale awkwardly
            self.text_entity.world_scale_x = self.text_entity.world_scale_y


if __name__ == '__main__':
    app = Ursina()

    health_bar_1 = HealthBar(bar_color=color.lime.tint(-.25), roundness=.5, value=50)

    def input(key):
        if key == '+' or key == '+ hold':
            health_bar_1.value += 10
        if key == '-' or key == '- hold':
            health_bar_1.value -= 10

    app.run()



import sys
import time
import traceback
from panda3d.core import *
from panda3d.core import CollisionTraverser, CollisionNode
from panda3d.core import CollisionHandlerQueue, CollisionRay
from ursina import application
from ursina.window import instance as window
from ursina.scene import instance as scene
from ursina.camera import instance as camera
from ursina.hit_info import HitInfo
from ursina.ursinamath import distance


class Mouse():

    def __init__(self):
        self.enabled = False
        self.visible = True
        self.locked = False
        self.position = Vec3(0,0,0)
        self.delta = Vec3(0,0,0)
        self.prev_x = 0
        self.prev_y = 0
        self.start_x = 0
        self.start_y = 0
        self.velocity = Vec3(0,0,0)
        self.prev_click_time = time.time()
        self.prev_click_pos = None
        self.double_click_distance = .5
        self.double_click_movement_limit = .01

        self.hovered_entity = None # returns the closest hovered entity with a collider.
        self.left = False
        self.right = False
        self.middle = False
        self.delta_drag = Vec3(0,0,0)

        self.update_step = 1
        self.traverse_target = scene  # set this to None to disable collision with scene, which might be a good idea if you have lots of colliders.
        self._i = 0
        self._mouse_watcher = None
        self._picker = CollisionTraverser()  # Make a traverser
        self._pq = CollisionHandlerQueue()  # Make a handler
        self._pickerNode = CollisionNode('mouseRay')
        self._pickerNP = camera.attach_new_node(self._pickerNode)
        self._pickerRay = CollisionRay()  # Make our ray
        self._pickerNode.addSolid(self._pickerRay)
        self._picker.addCollider(self._pickerNP, self._pq)
        self._pickerNode.set_into_collide_mask(0)

        self.raycast = True
        self.collision = None
        self.collisions = list()
        self.enabled = True

    @property
    def x(self):
        if not self._mouse_watcher.has_mouse():
            return 0
        return self._mouse_watcher.getMouseX() / 2 * window.aspect_ratio  # same space as ui stuff

    @x.setter
    def x(self, value):
        self.position = (value, self.y)


    @property
    def y(self):
        if not self._mouse_watcher.has_mouse():
            return 0

        return self._mouse_watcher.getMouseY() / 2

    @y.setter
    def y(self, value):
        self.position = (self.x, value)


    @property
    def position(self):
        return Vec3(self.x, self.y, 0)

    @position.setter
    def position(self, value):
        base.win.move_pointer(
            0,
            round(value[0] + (window.size[0]/2) + (value[0]/2*window.size[0]) *1.124), # no idea why I have * with 1.124
            round(value[1] + (window.size[1]/2) - (value[1]*window.size[1])),
        )



    def __setattr__(self, name, value):

        if name == 'visible':
            window.set_cursor_hidden(not value)
            if application.base:
                application.base.win.requestProperties(window)

        if name == 'locked':
            try:
                object.__setattr__(self, name, value)
                window.set_cursor_hidden(value)
                if value:
                    window.set_mouse_mode(window.M_relative)
                else:
                    window.set_mouse_mode(window.M_absolute)

                application.base.win.requestProperties(window)
            except:
                pass

        try:
            super().__setattr__(name, value)
            # return
        except:
            pass


    def input(self, key):
        if not self.enabled:
            return

        if key.endswith('mouse down'):
            self.start_x = self.x
            self.start_y = self.y

        elif key.endswith('mouse up'):
            self.delta_drag = Vec3(
                self.x - self.start_x,
                self.y - self.start_y,
                0
                )


        if key == 'left mouse down':
            self.left = True
            if self.hovered_entity:
                if hasattr(self.hovered_entity, 'on_click') and callable(self.hovered_entity.on_click):
                    try:
                        self.hovered_entity.on_click()
                    except Exception as e:
                        print(traceback.format_exc())
                        application.quit()

                for s in self.hovered_entity.scripts:
                    if hasattr(s, 'on_click') and callable(s.on_click):
                        s.on_click()

            # double click
            if time.time() - self.prev_click_time <= self.double_click_distance:
                if abs(self.x-self.prev_click_pos[0]) > self.double_click_movement_limit or abs(self.y-self.prev_click_pos[1]) > self.double_click_movement_limit:
                    return # moused moved too much since previous click, so don't register double click.

                base.input('double click')
                if self.hovered_entity:
                    if hasattr(self.hovered_entity, 'on_double_click'):
                        try:
                            self.hovered_entity.on_double_click()
                        except Exception as e:
                            print(traceback.format_exc())
                            application.quit()

                    for s in self.hovered_entity.scripts:
                        if hasattr(s, 'on_double_click'):
                            s.on_double_click()

            self.prev_click_time = time.time()
            self.prev_click_pos = (self.x, self.y)


        if key == 'left mouse up':
            self.left = False
        if key == 'right mouse down':
            self.right = True
        if key == 'right mouse up':
            self.right = False
        if key == 'middle mouse down':
            self.middle = True
        if key == 'middle mouse up':
            self.middle = False



    def update(self):
        if not self.enabled or not self._mouse_watcher.has_mouse():
            self.velocity = Vec3(0,0,0)
            return

        self.moving = self.x + self.y != self.prev_x + self.prev_y

        if self.moving:
            if self.locked:
                self.velocity = self.position
                self.position = (0,0)
            else:
                self.velocity = Vec3(self.x - self.prev_x, (self.y - self.prev_y) / window.aspect_ratio ,0)
        else:
            self.velocity = Vec3(0,0,0)

        if self.left or self.right or self.middle:
            self.delta = Vec3(self.x - self.start_x, self.y - self.start_y, 0)

        self.prev_x = self.x
        self.prev_y = self.y


        self._i += 1
        if self._i < self.update_step:
            return
        # collide with ui
        self._pickerNP.reparent_to(scene.ui_camera)
        self._pickerRay.set_from_lens(camera._ui_lens_node, self.x * 2 / window.aspect_ratio, self.y * 2)
        self._picker.traverse(camera.ui)
        if self._pq.get_num_entries() > 0:
            # print('collided with ui', self._pq.getNumEntries())
            self.find_collision()
            return

        # collide with world
        self._pickerNP.reparent_to(camera)
        self._pickerRay.set_from_lens(scene.camera.lens_node, self.x * 2 / window.aspect_ratio, self.y * 2)
        if self.traverse_target:
            self._picker.traverse(self.traverse_target)

        if self._pq.get_num_entries() > 0:
            self.find_collision()
        else:
            # print('mouse miss', base.render)
            # unhover all if it didn't hit anything
            for entity in scene.entities:
                if hasattr(entity, 'hovered') and entity.hovered:
                    entity.hovered = False
                    self.hovered_entity = None
                    if hasattr(entity, 'on_mouse_exit'):
                        entity.on_mouse_exit()
                    for s in entity.scripts:
                        if hasattr(s, 'on_mouse_exit'):
                            s.on_mouse_exit()

    @property
    def normal(self): # returns the normal of the polygon, in local space.
        if not self.collision is not None:
            return None
        return Vec3(*self.collision.normal)

    @property
    def world_normal(self): # returns the normal of the polygon, in world space.
        if not self.collision is not None:
            return None
        return Vec3(*self.collision.world_normal)

    @property
    def point(self): # returns the point hit, in local space
        if self.collision is not None:
            return Vec3(*self.collision.point)
        return None

    @property
    def world_point(self): # returns the point hit, in world space
        if self.collision is not None:
            return Vec3(*self.collision.world_point)
        return None

    def find_collision(self):
        self.collisions = list()
        self.collision = None
        if not self.raycast or self._pq.get_num_entries() == 0:
            self.unhover_everything_not_hit()
            return False

        self._pq.sortEntries()

        for entry in self._pq.getEntries():
            for entity in scene.entities:
                if entry.getIntoNodePath().parent == entity and entity.collision:
                    if entity.collision:
                        hit = HitInfo(
                            hit = entry.collided(),
                            entity = entity,
                            distance = distance(entry.getSurfacePoint(scene), camera.getPos()),
                            point = entry.getSurfacePoint(entity),
                            world_point = entry.getSurfacePoint(scene),
                            normal = entry.getSurfaceNormal(entity),
                            world_normal = entry.getSurfaceNormal(scene),
                            )
                        self.collisions.append(hit)
                        break

        if self.collisions:
            self.collision = self.collisions[0]
            self.hovered_entity = self.collision.entity
            if not self.hovered_entity.hovered:
                self.hovered_entity.hovered = True
                if hasattr(self.hovered_entity, 'on_mouse_enter'):
                    self.hovered_entity.on_mouse_enter()
                for s in self.hovered_entity.scripts:
                    if hasattr(s, 'on_mouse_enter'):
                        s.on_mouse_enter()


        self.unhover_everything_not_hit()



    def unhover_everything_not_hit(self):
        for e in scene.entities:
            if e == self.hovered_entity:
                continue

            if hasattr(e, 'hovered') and e.hovered:
                e.hovered = False
                if hasattr(e, 'on_mouse_exit'):
                    e.on_mouse_exit()
                for s in e.scripts:
                    if hasattr(s, 'on_mouse_exit'):
                        s.on_mouse_exit()



instance = Mouse()


if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    Button(parent=scene, text='a')

    def update():
        print(mouse.position, mouse.point)

    Cursor()
    mouse.visible = False


    app.run()



from ursina import *


class FirstPersonController(Entity):
    def __init__(self, **kwargs):
        self.cursor = Entity(parent=camera.ui, model='quad', color=color.pink, scale=.008, rotation_z=45)
        super().__init__()
        self.speed = 5
        self.height = 2
        self.camera_pivot = Entity(parent=self, y=self.height)

        camera.parent = self.camera_pivot
        camera.position = (0,0,0)
        camera.rotation = (0,0,0)
        camera.fov = 90
        mouse.locked = True
        self.mouse_sensitivity = Vec2(40, 40)

        self.gravity = 1
        self.grounded = False
        self.jump_height = 2
        self.jump_duration = .5
        self.jumping = False
        self.air_time = 0

        for key, value in kwargs.items():
            setattr(self, key ,value)


    def update(self):
        self.rotation_y += mouse.velocity[0] * self.mouse_sensitivity[1]

        self.camera_pivot.rotation_x -= mouse.velocity[1] * self.mouse_sensitivity[0]
        self.camera_pivot.rotation_x= clamp(self.camera_pivot.rotation_x, -90, 90)

        self.direction = Vec3(
            self.forward * (held_keys['w'] - held_keys['s'])
            + self.right * (held_keys['d'] - held_keys['a'])
            ).normalized()

        feet_ray = raycast(self.position+Vec3(0,0.5,0), self.direction, ignore=(self,), distance=.5, debug=False)
        head_ray = raycast(self.position+Vec3(0,self.height-.1,0), self.direction, ignore=(self,), distance=.5, debug=False)
        if not feet_ray.hit and not head_ray.hit:
            self.position += self.direction * self.speed * time.dt


        if self.gravity:
            # gravity
            ray = raycast(self.world_position+(0,self.height,0), self.down, ignore=(self,))
            # ray = boxcast(self.world_position+(0,2,0), self.down, ignore=(self,))

            if ray.distance <= self.height+.1:
                if not self.grounded:
                    self.land()
                self.grounded = True
                # make sure it's not a wall and that the point is not too far up
                if ray.world_normal.y > .7 and ray.world_point.y - self.world_y < .5: # walk up slope
                    self.y = ray.world_point[1]
                return
            else:
                self.grounded = False

            # if not on ground and not on way up in jump, fall
            self.y -= min(self.air_time, ray.distance-.05) * time.dt * 100
            self.air_time += time.dt * .25 * self.gravity


    def input(self, key):
        if key == 'space':
            self.jump()


    def jump(self):
        if not self.grounded:
            return

        self.grounded = False
        self.animate_y(self.y+self.jump_height, self.jump_duration, resolution=int(1//time.dt), curve=curve.out_expo)
        invoke(self.start_fall, delay=self.jump_duration)


    def start_fall(self):
        self.y_animator.pause()
        self.jumping = False

    def land(self):
        # print('land')
        self.air_time = 0
        self.grounded = True


    def on_enable(self):
        mouse.locked = True
        self.cursor.enabled = True


    def on_disable(self):
        mouse.locked = False
        self.cursor.enabled = False




if __name__ == '__main__':
    from ursina.prefabs.first_person_controller import FirstPersonController
    # window.vsync = False
    app = Ursina()
    # Sky(color=color.gray)
    ground = Entity(model='plane', scale=(100,1,100), color=color.yellow.tint(-.2), texture='white_cube', texture_scale=(100,100), collider='box')
    e = Entity(model='cube', scale=(1,5,10), x=2, y=.01, rotation_y=45, collider='box', texture='white_cube')
    e.texture_scale = (e.scale_z, e.scale_y)
    e = Entity(model='cube', scale=(1,5,10), x=-2, y=.01, collider='box', texture='white_cube')
    e.texture_scale = (e.scale_z, e.scale_y)

    player = FirstPersonController(y=2, origin_y=-.5)
    player.gun = None


    gun = Button(parent=scene, model='cube', color=color.blue, origin_y=-.5, position=(3,0,3), collider='box')
    gun.on_click = Sequence(Func(setattr, gun, 'parent', camera), Func(setattr, player, 'gun', gun))

    gun_2 = duplicate(gun, z=7, x=8)
    slope = Entity(model='cube', collider='box', position=(0,0,8), scale=6, rotation=(45,0,0), texture='brick', texture_scale=(8,8))
    slope = Entity(model='cube', collider='box', position=(5,0,10), scale=6, rotation=(80,0,0), texture='brick', texture_scale=(8,8))
    # hill = Entity(model='sphere', position=(20,-10,10), scale=(25,25,25), collider='sphere', color=color.green)
    # hill = Entity(model='sphere', position=(20,-0,10), scale=(25,25,25), collider='mesh', color=color.green)
    # from ursina.shaders import basic_lighting_shader
    # for e in scene.entities:
    #     e.shader = basic_lighting_shader

    hookshot_target = Button(parent=scene, model='cube', color=color.brown, position=(4,5,5))
    hookshot_target.on_click = Func(player.animate_position, hookshot_target.position, duration=.5, curve=curve.linear)

    def input(key):
        if key == 'left mouse down' and player.gun:
            gun.blink(color.orange)
            bullet = Entity(parent=gun, model='cube', scale=.1, color=color.black)
            bullet.world_parent = scene
            bullet.animate_position(bullet.position+(bullet.forward*50), curve=curve.linear, duration=1)
            destroy(bullet, delay=1)

    # player.add_script(NoclipMode())
    app.run()
